import math
import itertools
from collections import OrderedDict

import torch
import torch.nn as nn
from torch.onnx import is_in_onnx_export

from ..config import config
from ..quant import (Quant, normalize_bias_with_config,
                     normalize_data_with_config, normalize_weight_with_config)
from ..utils import Dump, PlatFormQuant, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear)
from .ops import ModuleIntConfig
from .requant import Requant

import lingerext

class LayerNormFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weights, bias, normalized_shape,
                training, momentum, eps,
                running_x, running_normal, running_w, running_o,
                eval_scale_x, eval_scale_normal, eval_scale_w, eval_scale_o,
                data_bits, parameter_bits, prefix, dump,
                path, mode, o_bits, quant, is_not_from_iqtensor,
                ahead_relu, clamp_data, clamp_weight, clamp_bias):
        N = 1
        for dim_size in normalized_shape:
            N *= dim_size
        if len(weights.shape) == 1:
            w_shape = (-1)
        elif len(weights.shape) == 2:
            w_shape = (-2, -1)
        elif len(weights.shape) == 3:
            w_shape = (-3, -2, -1)
        elif len(weights.shape) == 4:
            w_shape = (-4, -3, -2, -1)
        else:
            assert False, f"weight.shape=={weights.shape}, please check your LayerNorm definition."
        H, W = input.shape[-2], input.shape[-1]
        scale_x = None
        scale_w = None
        scale_o = None
        if training:
            ctx.clamp_data = clamp_data
            ctx.clamp_weight = clamp_weight
            ctx.clamp_bias = clamp_bias
            ctx.eps = eps
            ctx.bits = data_bits, parameter_bits, o_bits
            zero_point = input.zero_point if isinstance(input, IQTensor) else 0
            is_iq_tensor = True if isinstance(input, IQTensor) else False
            ctx.value = zero_point, is_iq_tensor
            ctx.N = N
            ctx.w_shape = w_shape
            saved_tensors = [input, weights, bias]

            # x
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(
                    input.data, data_bits, eval_scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
                scale_x = eval_scale_x
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                running_x.mul_(1-momentum).add_(momentum*max_value_x)
            # q_input = q_input.float() if data_bits + parameter_bits <= 16 else q_input.double()
            q_input = q_input.contiguous().int()
            sum_x = q_input.clone().sum(w_shape, keepdim=True)
            sum_x2 = q_input.clone().pow(2).sum(w_shape, keepdim=True)
            denominator = N * sum_x2 - sum_x * sum_x
            denominator = denominator.contiguous().int()
            denominator, deno_scale= lingerext.luna_layernormint(denominator.contiguous(), math.log2(scale_x.data))
            deno_scale_f = torch.pow(2, deno_scale.float())
            q_x_normal = q_input * N
            q_x_normal = q_x_normal - sum_x
            q_x_normal = q_x_normal * denominator
            q_x_normal = (q_x_normal * deno_scale_f + 0.5).floor().int()
            q_x_normal.clamp_(-2**15, 2**15-1)
            eval_scale_normal.fill_(2**12)

            q_weights, scale_w, max_value_w = quant.quant(
                weights, parameter_bits, mode=mode, quant_data='weight')
            running_w.mul_(1-momentum).add_(momentum*max_value_w)

            q_outputs = q_x_normal * q_weights

            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                q_bias = (bias * scale_w * eval_scale_normal + 0.5).floor()
                q_bias.clamp_(-2**31, 2**31-1)
                q_outputs = q_outputs + q_bias
                q_outputs.clamp_(-2**31, 2**31-1)
                outputs = quant.dequant(q_outputs, eval_scale_normal*scale_w)
            else:
                assert False, "linger only support luna quant."

            out_tensor = outputs
            scale_o = None
            if o_bits is not None:
                outputs = normalize_data_with_config(outputs, clamp_data)
                out_tensor = outputs
                q_outputs, scale_o, max_value_o = quant.quant(
                    outputs, o_bits, mode=mode, quant_data='output', ahead_relu=ahead_relu)
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
                outputs = quant.dequant(q_outputs, scale_o)
            saved_tensors += [out_tensor]
            ctx.scale = scale_x, eval_scale_normal, scale_w, scale_o
            ctx.save_for_backward(*saved_tensors)

        else:
            assert running_x > 0, 'invalid running_x <= 0, please finetune training before eval'
            eval_scale_normal.fill_(2**12)
            if weights.dtype == torch.float32:
                scale_x = ScalerBuffer(quant.running_to_scale(
                    running_x, data_bits, mode=mode))
                if o_bits is not None:
                    scale_o = ScalerBuffer(quant.running_to_scale(
                        running_o, o_bits, mode=mode))
                weigths = normalize_weight_with_config(
                    weights, clamp_weight, False)
                bias = normalize_bias_with_config(bias, clamp_bias, False)
                q_weights, scale_w, _ = quant.quant(
                    weigths, parameter_bits, mode=mode, quant_data='weight')
                q_bias = None
                if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                    if bias.dtype == torch.float32:
                        q_bias = (bias * scale_w * eval_scale_normal + 0.5).floor()
                    else:
                        q_bias = bias.contiguous()                    
                else:
                    assert False, 'linger only support luna quant.'
            else:
                scale_x = eval_scale_x
                scale_normal = eval_scale_normal
                scale_w = eval_scale_w
                scale_o = eval_scale_o
                q_weights = weights.contiguous()
                q_bias = bias.contiguous()
            q_bias.clamp_(-2**31, 2**31-1)
            if isinstance(input, IQTensor):
                scale_x = eval_scale_x
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=mode, quant_data='input')
            q_input = q_input.contiguous().int()
            q_weights = q_weights.contiguous().int()
            sum_x = q_input.clone().sum(w_shape, keepdim=True)
            sum_x2 = q_input.clone().pow(2).sum(w_shape, keepdim=True)
            denominator = N * sum_x2 - sum_x * sum_x
            denominator = denominator.contiguous().int()
            denominator, deno_scale= lingerext.luna_layernormint(denominator.contiguous(), math.log2(scale_x.data))
            deno_scale_f = torch.pow(2, deno_scale.float())
            q_x_normal = q_input * N
            q_x_normal = q_x_normal - sum_x
            q_x_normal = q_x_normal * denominator
            q_x_normal = (q_x_normal * deno_scale_f + 0.5).floor().int()
            q_x_normal.clamp_(-2**15, 2**15-1)
            q_outputs = q_x_normal * q_weights + q_bias
            q_outputs.clamp_(-2**31, 2**31-1)
            outputs = quant.dequant(q_outputs, eval_scale_normal * scale_w)

            if o_bits is not None:
                if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                    q_outputs, _, _ = quant.quant(
                        outputs, o_bits, scale_o, mode=mode, quant_data='output')
                else:
                    assert False, "linger only support luna quant."
                outputs = quant.dequant(q_outputs, scale_o)

            if dump:
                name_list = ["input", "weights", "bias", "q_weights",  "q_input", "q_input_normal", "denominator", "q_bias", "q_outputs",
                             "scale_normal", "scale_x", "scle_w", "scale_o"]
                attr_list = [input, weights, bias, q_weights, q_input, q_x_normal, denominator, q_bias, q_outputs,
                             eval_scale_normal.data, scale_x.data, scale_w.data, scale_o.data]
                Dump.dump_file(prefix, ".LayerNormInt.",
                               zip(name_list, attr_list), path)

        if isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradOutput):
        input, weights, bias, outputs = ctx.saved_tensors
        clamp_data = ctx.clamp_data
        data_bits, parameter_bits, o_bits = ctx.bits
        scale_x, scale_normal, scale_w, scale_o = ctx.scale
        zero_point, is_iq_tensor = ctx.value
        N = ctx.N
        w_shape = ctx.w_shape
        eps = ctx.eps

        if is_iq_tensor:
            f_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            f_input = Quant.dequant(q_input, scale_x)

        f_input = f_input.detach().clone().requires_grad_(True)

        q_weights, _, _ = Quant.quant(
            weights.data, parameter_bits, scale_w, mode=QuantMode.QValue, quant_data='weight')
        f_weights = Quant.dequant(q_weights, scale_w)
        f_weights = f_weights.detach().clone().requires_grad_(True)
        bias = None if bias is None else bias.detach().clone().requires_grad_(True)

        with torch.enable_grad():
            mean = f_input.clone().sum(w_shape, keepdim=True) / N
            var = input.clone().pow(2).sum(w_shape, keepdim=True) / N - \
                (input.clone().sum(w_shape, keepdim=True) / N).pow(2)
            var = 1/torch.sqrt(var + eps)
            var = torch.clamp(var, min=0.0)
            f_input_normal = (input - mean) * var
            z = f_weights * f_input_normal + bias
            z = normalize_data_with_config(z, clamp_data)
            gradInput, gradWeight, gradBias = torch.autograd.grad(
                z, (f_input, f_weights, bias), gradOutput)

        return gradInput, gradWeight, gradBias, None, \
            None, None, None, None, None, \
            None, None, None, \
            None, None, None, None, \
            None, None, None, None, \
            None, None, None, None, \
            None, None, None, None, None, \
            None, None, None, None

    @staticmethod
    def symbolic(g, input, weights, bias, normalized_shape,
                 training, momentum, eps,
                 running_x, running_normal, running_w, running_o,
                 scale_x, scale_normal, scale_w, scale_o,
                 data_bits, parameter_bits, prefix, dump,
                 path, mode, o_bits, quant, is_not_from_iqtensor,
                 ahead_relu, clamp_data, clamp_weight, clamp_bias):
        op_inner = None
        if is_not_from_iqtensor:
            platform_quant = platform_to_string(
                config.PlatFormQuant.platform_quant)
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
        param_dict = {'data_bits_i': data_bits, 'parameter_bits_i': parameter_bits, 'o_bits_i': o_bits,
                      'scale_x_f': scale_x(), 'scale_w_f': scale_w(), 'scale_o_f': scale_o()}

        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant
        if is_not_from_iqtensor:
            input_list = [op_inner, weights, bias]
        else:
            input_list = [input, weights, bias]
        return g.op("thinker::LayerNormInt", *input_list, **param_dict)


class LayerNormInt(nn.LayerNorm, ModuleIntConfig):
    r"""实现LayerNormInt的量化训练与测试，继承自nn.LayerNorm,

    Args:
        num_features eps momentum affine track_running_stats
        标准nn.LayerNorm的参数
        data_bits(int): 输入量化位数
        parameter_bits(int): 参数量化位数
        mode(Enum): 量化方式，支持MaxValue与Qvalue
        o_bits(int, default=None):输出量化位数
        clamp_data(float or None): 针对输出的clamp数值
        clamp_weight(float or None): 针对转为乘加操作之后的weight与bias的clamp数值，此处不使用
        clamp_bias(float or None): 与clamp_weight一致
        ahead_relu(bool): 是否做融合relu之后的数值统计scale

    Examples:
        test/test_layernorm_int.py

    """

    def __init__(self, normalized_shape, eps=1e-05, momentum=0.1, elementwise_affine=True, data_bits=8, parameter_bits=16, mode=QuantMode.QValue, o_bits=8,
                 clamp_data=None, clamp_weight=None, clamp_bias=None, ahead_relu=False):
        parameter_bits = 16
        nn.LayerNorm.__init__(self, normalized_shape, eps, elementwise_affine)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=o_bits)
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.is_not_from_iqtensor = True
        self.ahead_relu = ahead_relu
        self.clamp_data = clamp_data
        self.clamp_weight = clamp_weight
        self.clamp_bias = clamp_bias
        self.mode = mode
        self.momentum = momentum
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_normal', torch.zeros(1))
        self.register_buffer('running_w', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_normal', torch.zeros(1))
        self.register_buffer('scale_w', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input):
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits, self.mode)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        running_normal = ScalerBuffer(self.running_normal)
        running_w = ScalerBuffer(self.running_w)
        running_o = ScalerBuffer(self.running_o)
        scale_normal = ScalerBuffer(self.scale_normal)
        scale_w = ScalerBuffer(self.scale_w)
        scale_o = ScalerBuffer(self.scale_o)
        momentum = 0.1
        weight = self.weight
        bias = self.bias
        if self.weight.dtype == torch.float32:
            weight = normalize_weight_with_config(
                self.weight, self.clamp_weight, self.training)
            if self.bias is not None:
                bias = normalize_bias_with_config(
                    self.bias, self.clamp_bias, self.training)
        
        ret = LayerNormFunction.apply(input, weight, bias, self.normalized_shape,
                                      self.training and self.elementwise_affine, momentum, self.eps,
                                      running_x, running_normal, running_w, running_o,
                                      scale_x, scale_normal, scale_w, scale_o,
                                      self.data_bits, self.parameter_bits, self.prefix, self.dump,
                                      self.path, self.quant_mode, self.o_bits, self.quant, self.is_not_from_iqtensor,
                                      self.ahead_relu, self.clamp_data, self.clamp_weight, self.clamp_bias)
        self.running_x.fill_(running_x())
        self.running_normal.fill_(running_normal())
        self.running_w.fill_(running_w())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_normal.fill_(scale_normal())
        self.scale_w.fill_(scale_w())
        self.scale_o.fill_(scale_o())
        return ret

    def state_dict(module, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=module._version)
        scale_x = ScalerBuffer(module._buffers['scale_x'])
        scale_normal = ScalerBuffer(module._buffers['scale_normal'])
        if is_in_onnx_export():
            assert module._buffers['running_x'] > 0, 'invalid running_x <= 0, cannot access param before training, layer prefix is: {}'.format(
                prefix)
            if module.is_not_from_iqtensor:
                scale_x = ScalerBuffer(module.quant.running_to_scale(ScalerBuffer(
                    module._buffers['running_x']), module.data_bits, mode=module.quant_mode))
                module._buffers['scale_x'].data.fill_(scale_x())
            # scale_normal = ScalerBuffer(module.quant.running_to_scale(ScalerBuffer(
            #     module._buffers['running_normal']), module.data_bits, mode=module.quant_mode))
            scale_normal.fill_(2**12)
            module._buffers['scale_normal'].data.fill_(scale_normal())
            if module.o_bits is not None:
                scale_o = ScalerBuffer(module.quant.running_to_scale(ScalerBuffer(
                    module._buffers['running_o']), module.o_bits, mode=module.quant_mode))
                module._buffers['scale_o'].data.fill_(scale_o())
        if 'scale_w' in module._buffers and module._parameters['weight'].dtype == torch.float:
            weight_tensor = module._parameters['weight']
            weight_tensor_clamp = None
            bias_tensor_clamp = None
            if hasattr(module, 'clamp_weight'):
                weight_tensor_clamp = normalize_weight_with_config(
                    weight_tensor, module.clamp_weight, False)
            else:
                weight_tensor_clamp = weight_tensor
            weight_tensor.data = weight_tensor_clamp
            if module.bias is not None:
                bias_tensor = module._parameters['bias']
                # bias_temp = None
                if hasattr(module, 'clamp_bias'):
                    bias_tensor_clamp = normalize_bias_with_config(
                        bias_tensor, module.clamp_bias, False)
                else:
                    bias_tensor_clamp = bias_tensor
                bias_tensor.data = bias_tensor_clamp
            if is_in_onnx_export():
                weight_temp, scale_w, _ = module.quant.quant(
                    weight_tensor_clamp, module.parameter_bits, mode=module.quant_mode)
                scale_w = ScalerBuffer(scale_w)
                module._buffers['scale_w'].data.fill_(scale_w())

                if module.parameter_bits <= 8:
                    weight_tensor.data = weight_temp.char()
                    weight_tensor.char()
                elif module.parameter_bits <= 16:
                    weight_tensor.data = weight_temp.short()
                    weight_tensor.short()
                else:
                    weight_tensor.data = weight_temp.int()
                    weight_tensor.int()
                if module.bias is not None:
                    bias_tensor_clamp = module._parameters['bias']
                    if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                        assert module.quant_mode == QuantMode.QValue, 'luna_quant only support Qvalue and o_bits=None'
                        if module.data_bits + module.parameter_bits <= 16:
                            module._parameters['bias'].data = (
                                bias_tensor_clamp * scale_w * scale_normal + 0.5).floor().float().int()
                        else:
                            module._parameters['bias'].data = (
                                bias_tensor_clamp * scale_w * scale_normal + 0.5).floor().int()
                        module._parameters['bias'].int()
                    else:
                        assert False, "linger only support luna quant."
        module._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in module._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in module._state_dict_hooks.values():
            hook_result = hook(module, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def _load_from_state_dict(module, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        allow_missing_keys = ['running_w', 'running_x', 'running_normal', 'running_o',
                              'scale_x', 'running_normal', 'scale_w', 'scale_o']
        local_missing_keys = []
        module._load_from_state_dict_global_(state_dict, prefix, local_metadata, strict,
                                             local_missing_keys, unexpected_keys, error_msgs)
        matched = True
        fake_missing_keys = []
        for k_local in local_missing_keys:
            if k_local.replace(prefix, '', 1) not in allow_missing_keys:
                matched = False
                fake_missing_keys.append(k_local)
        if matched:
            local_missing_keys = []
        else:
            local_missing_keys = fake_missing_keys
        missing_keys += local_missing_keys

    def _load_from_state_dict_global_(module, state_dict, prefix, local_metadata, strict,
                                      missing_keys, unexpected_keys, error_msgs):
        for hook in module._load_state_dict_pre_hooks.values():
            hook(state_dict, prefix, local_metadata, strict,
                 missing_keys, unexpected_keys, error_msgs)
        local_name_params = itertools.chain(
            module._parameters.items(), module._buffers.items())
        local_state = {k: v.data for k,
                       v in local_name_params if v is not None}
        for name, param in local_state.items():
            key = prefix + name
            if key in state_dict:
                input_param = state_dict[key]

                if len(param.shape) == 0 and len(input_param.shape) == 1:
                    input_param = input_param[0]

                if input_param.shape != param.shape:
                    error_msgs.append('size mismatch for {}: copying a param with shape {} from checkpoint, '
                                      'the shape in current model is {}.'
                                      .format(key, input_param.shape, param.shape))
                    continue

                if isinstance(input_param, torch.nn.Parameter):
                    input_param = input_param.data
                try:
                    param.copy_(input_param)
                    if input_param.dtype == torch.int32 or input_param.dtype == torch.int8 or input_param.dtype == torch.int16:
                        module._parameters[name] = param.int()

                except Exception:
                    error_msgs.append('While copying the parameter named "{}", '
                                      'whose dimensions in the model are {} and '
                                      'whose dimensions in the checkpoint are {}.'
                                      .format(key, param.size(), input_param.size()))
            elif strict:
                missing_keys.append(key)
        if strict:
            for key in state_dict.keys():
                if key.startswith(prefix):
                    input_name = key[len(prefix):]
                    input_name = input_name.split('.', 1)[0]
                    if input_name not in module._modules and input_name not in local_state:
                        unexpected_keys.append(key)

    def extra_repr(self):
        s = nn.LayerNorm.extra_repr(self)
        extra_s = ' ,clamp_data:{clamp_data},clamp_weight:{clamp_weight},clamp_bias:{clamp_bias},ahead_relu:{ahead_relu}'.format(
            **self.__dict__)
        extra_s += ',data_bits:{data_bits}, parameter_bits:{parameter_bits}, o_bits:{o_bits}'.format(
            **self.__dict__)
        return s+extra_s



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import torch
import torch.nn as nn
import linger
import numpy as np

def test_s_relu6_int():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                         padding=1, bias=True)
            self.bn = nn.BatchNorm2d(10)
            self.relu = nn.ReLU6()
            self.conv1 = nn.Conv2d(10, 10, kernel_size=1, stride=1,
                         padding=0, bias=True)
            self.fc = nn.Linear(1000, 100)

        def forward(self, x):
            x = self.conv(x)
            x = self.bn(x)
            # x = x+8
            x = self.relu(x)
            x = self.conv1(x)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))  
            x = self.fc(x)
            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    # random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    aa = torch.randn(1, 10, 10, 10).cuda()
    # import pickle as pk
    # with open('input.pb', 'wb') as file:
    #     pk.dump(aa, file)
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    # replace_tuple=(nn.Conv2d, nn.Linear, nn.AvgPool2d, nn.ReLU6)
    replace_tuple=(nn.ReLU6)
    net.train()
    # linger.disable_quant(net.fc)
    net = linger.init(net, quant_modules=replace_tuple, mode=linger.QuantMode.QValue)
    optimizer = torch.optim.SGD(net.parameters(), lr = 0.01)
    loss = None

    for i in range(200):
        optimizer.zero_grad()
        out = net(aa)
        loss = criterion(out, target)
        if i % 20 == 0:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()
    # assert loss < 1e-12, 'training loss error'
    net.eval()
    torch.save(net.state_dict(), 'data.ignore/relu6_int.pt')
    net.load_state_dict(torch.load('data.ignore/relu6_int.pt'))
    out1 = net(aa)
    # print(out1)
    with torch.no_grad():
        torch.onnx.export(net, aa,"data.ignore/test_relu6_int.onnx",export_params=True,opset_version=12,operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    assert abs(out1.mean() - 1) < 0.01



import torch.nn as nn

from .conv_bn_fuser import *


def trace_layers(root_model: nn.Module, target_model: nn.Module, *args, fuse_bn: bool = True, ahead_conv_relu: bool = True, ahead_bn_relu: bool = True, ahead_linear_relu: bool = True, ahead_conv_sigmoid: bool = True, ahead_linear_sigmoid: bool = True) -> nn.Module:
    r"""对模型进行trace 同时进行模型提前进行fusion训练,root_model为原始的根module,target_model为目标trace的子model

    Args:
        root_model(torch.nn.Module): 原始根模型
        target_model(torch.nn.Module): Trace 子模型
        args(torch.Tensor or Tuple or List): 模型Trace的位置参数
        fuse_bn(bool)：是否融合BN 
        ahead_conv_relu(bool): 是否统计Conv输出正值scale
        ahead_bn_relu(bool): 是否统计BN输出正值scale 
        ahead_linear_relu(bool): 是否统计Linear输出正值scale
    returns:
        返回融合BN后的module

    Examples:
        test/test_trace_layers.py

    Note:
        trace_layer 只支持一次调用，对于多次调用时，会直接覆盖前一次trace_layer的实现，但并未体现在输出的net的网络结构中，内部hook会被清空，导致加载的融合参数无法生效

    """
    if SingletonConvFusedBnModules().has_fuseable_items():
        print("Warning: trace_layers only support one-time call, the latest call will overwrite the previous call and this may cause errors, please check !")
    FuseConvBNAheadRelu(target_model, *args, fused_bn=fuse_bn, ahead_conv_relu=ahead_conv_relu, ahead_bn_relu=ahead_bn_relu,
                        ahead_linear_relu=ahead_linear_relu, ahead_conv_sigmoid=ahead_conv_sigmoid, ahead_linear_sigmoid=ahead_linear_sigmoid)
    if SingletonConvFusedBnModules().has_fuseable_items():
        SingletonConvFusedBnModules().build_normalize_convbn2d_scope(root_model)
        SingletonConvFusedBnModules().build_empty_bn_scope(root_model)

    def pre_hook(state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
        SingletonConvFusedBnModules().fuse_state_dicts(state_dict)
        SingletonConvFusedBnModules().clear()
    if SingletonConvFusedBnModules().has_fuseable_items():
        root_model._register_load_state_dict_pre_hook(pre_hook)
    return root_model



import torch.nn as nn

from ..modules import *
from .avgpool2d_int import AvgPool2dInt
from .batchnorm_int import BatchNormInt
from .bmm_int import BmmInt
from .conv1d_int import Conv1dInt
from .conv_int import Conv2dInt
from .convtranspose_int import ConvTranspose2dInt
from .embedding_int import EmbeddingInt
from .gru_int import GRUInt
from .iqtensor import iqAddLayer, iqDivLayer, iqMulLayer, iqSumLayer
from .linear_int import LinearInt
from .linger_functional import (iqCatLayer, iqClampLayer, iqSigmoidLayer,
                                iqTanhLayer, softmaxInt, logsoftmaxInt)
from .lstm_int import LSTMInt
from .relu6_int import ReLU6Int
from .layernorm_int import LayerNormInt

DefaultQuantIntXOP = (nn.BatchNorm2d, nn.Linear, nn.Conv2d, nn.ConvTranspose2d, nn.AvgPool2d, nn.Conv1d, NormalizeConvBN1d, NormalizeConvBN2d,
                      NormalizeConv2d, NormalizeConv1d, NormalizeConvTranspose2d, NormalizeLinear, NormalizeFastLSTM, NormalizeFastGRU,nn.ReLU6, nn.GRU, nn.LSTM)

SupportQuantTorchModules = [nn.BatchNorm2d, nn.Linear, nn.Conv2d, nn.ConvTranspose2d, nn.GRU, nn.LSTM, nn.AvgPool2d, nn.Conv1d,  NormalizeConvBN1d, NormalizeConvBN2d, NormalizeConv2d,
                            NormalizeConv1d, NormalizeConvTranspose2d, NormalizeLinear, nn.ReLU6, NormalizeFastLSTM, NormalizeFastGRU, NormalizeBatchNorm2d, NormalizeLayerNorm, nn.Embedding, nn.Upsample, NormalizeEmbedding, nn.LayerNorm]
SupportQuantedIntModules = (BatchNormInt, LinearInt, Conv2dInt, ConvTranspose2dInt, GRUInt, LSTMInt, AvgPool2dInt, Conv1dInt, ReLU6Int, BmmInt,
                            iqAddLayer, iqMulLayer, iqDivLayer, iqSumLayer, iqCatLayer, iqSigmoidLayer, iqTanhLayer, iqClampLayer, EmbeddingInt, LayerNormInt, softmaxInt, logsoftmaxInt)



import torch
import prettytable as pt


def clamp_with_dynamic(input: torch.Tensor, dynamic_percent: float = 0.9, layer_name: str = "layer", tb_all=None):
    clamp_value = 0
    if dynamic_percent < 1.0:
        x = input.data.clone().cpu()
        x = x.abs().reshape(-1).sort().values
        clamp_index = int(len(x) * dynamic_percent) - 1
        clamp_value = x[clamp_index]
        max_abs = x[-1]
        if len(x) == 1:
            mean_abs = x[-1]
        else:
            mean_abs = x.mean()
        length = max_abs / mean_abs
        versu = max_abs / clamp_value
        if length >= 1:
            index_list = [layer_name, mean_abs,
                          max_abs, length, clamp_value, versu]
            tb_all.add_row(index_list)
        if versu > 10:
            index_list = ["!!!!!!!!!!!!!!", "!!!!!!!!!!!!!!", "!!!!!!!!!!!!!!",
                          "!!!!!!!!!!!!!!", "!!!!!!!!!!!!!!", "!!!!!!!!!!!!!!"]
            tb_all.add_row(index_list)


def wb_analyse(path: str , save_log_path: str ="wb_analyse.log"):
    if isinstance(path, str):
        model = torch.load(path)
    else:
        model = path

    wb_flile_path = save_log_path
    wb_flile = open(wb_flile_path, 'w+')
    tb_all = pt.PrettyTable()

    tb_all.field_names = ["Layer_name", "Mean", "Max",
                          "Multiple(Max/Mean)", "Dynamic 0.99", "Versu(Max/Dynamic)"]

    for k in model.keys():
        if "running" not in k:
            v = model[k]
            clamp_with_dynamic(v, 0.99, k, tb_all)

    wb_flile.write(str(tb_all))
    wb_flile.close()



import os
import re

import numpy as np
import prettytable as pt
import torch
import torch.nn as nn

from .ops import *
from .tools.weight_bias_analyse import clamp_with_dynamic

tb_all = pt.PrettyTable()

tb_all.field_names = ["Layer_name", "Mean", "Max",
                      "Multiple(Max/Mean)", "Dynamic 0.99", "Versu(Max/Dynamic)"]


def _hook_forward_anylse(module, input, output):
    if 1:  # not module.training:
        assert hasattr(module, LINGER_DUMP_NAME)
        file_path = getattr(module, LINGER_DUMP_NAME)
        if type(output) is tuple:
            if isinstance(module, nn.GRU) or isinstance(module, nn.LSTM):
                if isinstance(output[0], torch.nn.utils.rnn.PackedSequence):
                    dump_out = output[0][0]
                else:
                    dump_out = output[0]
            else:
                if type(output[1]) is tuple:
                    dump_out = torch.cat(
                        (output[0], output[1][0], output[1][1]))
                else:
                    dump_out = torch.cat(output)
        else:
            dump_out = output

        file_path = file_path[5:]  # remove "root."

        clamp_with_dynamic(dump_out.detach(), dynamic_percent=0.9,
                    layer_name=file_path, tb_all=tb_all)


def _hook_forward(module, input, output):
    if not module.training:
        assert hasattr(module, LINGER_DUMP_NAME)
        file_path = getattr(module, LINGER_DUMP_NAME)
        if type(input) is tuple:
            if len(input) == 1:
                dump_in = input[0]  # 单一输入 在此处也是tuple类型
            else:
                dump_in = input[0]  # 多个输入  只保存第一个输入
        else:
            dump_in = input
        if type(output) is tuple:
            if isinstance(module, nn.GRU) or isinstance(module, nn.LSTM):
                if isinstance(output[0], torch.nn.utils.rnn.PackedSequence):
                    dump_out = output[0][0]
                else:
                    dump_out = output[0]
            else:
                if type(output[1]) is tuple:
                    dump_out = torch.cat(
                        (output[0], output[1][0], output[1][1]))
                else:
                    dump_out = torch.cat(output)
        else:
            dump_out = output
        np.savetxt(file_path+'_input_float',
                   dump_in.detach().reshape(-1).cpu().numpy(), fmt='%f')
        np.savetxt(file_path+'_output_float',
                   dump_out.detach().reshape(-1).cpu().numpy(), fmt='%f')


def _dfs_submodules(model):
    dfs_modules = []
    stack = [('root', model)]
    while len(stack) > 0:
        (name_m, m) = stack.pop()
        children_num = 0
        for name, submodule in m.named_children():
            stack.append((name_m+'/'+name, submodule))
            children_num += 1
        if children_num == 0:
            dfs_modules.append((name_m, m))
    dfs_modules.reverse()
    return dfs_modules


class Dumper():
    def __init__(self):
        self.module_dump_quanted = []
        self.module_hooks_dump_all = []

    def __enter__(self):
        self.module_dump_quanted = []
        self.module_hooks_dump_all = []
        return self

    def __exit__(self, type, value, trace):
        self._clear_dump_quanted()
        self._clear_dump_model()

    def enable_dump_quanted(self, model: nn.Module, path: str = "./dump", match_pattern: str = ".*"):
        if model.training:
            print("error model must be eval when dump,call model.eval() before dump")
            exit(-1)
        if not os.path.exists(path):
            os.makedirs(path)
        queue = [("root", model)]
        while len(queue) > 0:
            name_m, node = queue.pop(0)
            for name, submodule in node.named_children():
                prefix = name_m + '.' + name
                queue.append((prefix, submodule))
                if isinstance(submodule, SupportQuantedIntModules) and re.match(match_pattern, prefix) is not None:
                    submodule.prefix = prefix
                    submodule.dump = True
                    submodule.path = path
                    self.module_dump_quanted.append(submodule)

    def _clear_dump_quanted(self):
        for op in self.module_dump_quanted:
            op.prefix = ''
            op.dump = False
            op.path = None
        self.module_dump_quanted = []

    def enable_dump_model(self, model: nn.Module, path: str ="./dump", match_pattern: str =".*", hook_forward=_hook_forward):
        if model.training:
            print("error model must be eval when dump,call model.eval() before dump")
            exit(-1)
        if not os.path.exists(path):
            os.makedirs(path)
        leaf_all_modules = _dfs_submodules(model)
        for name, leaf_module in leaf_all_modules:
            dump_name = name.replace('/', '.')
            if re.match(match_pattern, dump_name) is not None:
                setattr(leaf_module, LINGER_DUMP_NAME,
                        os.path.join(path, dump_name))
                hook_handle = leaf_module.register_forward_hook(_hook_forward)
                self.module_hooks_dump_all.append((leaf_module, hook_handle))

    def _clear_dump_model(self):
        for (m, hook) in self.module_hooks_dump_all:
            hook.remove()
            delattr(m, LINGER_DUMP_NAME)
        self.module_hooks_dump_all = []

    def analyse_layer_output(self, model: nn.Module, match_pattern: str =".*"):
        leaf_all_modules = _dfs_submodules(model)

        for name, leaf_module in leaf_all_modules:
            dump_name = name.replace('/', '.')
            if re.match(match_pattern, dump_name) is not None:
                setattr(leaf_module, LINGER_DUMP_NAME, dump_name)
                hook_handle = leaf_module.register_forward_hook(
                    _hook_forward_anylse)
                self.module_hooks_dump_all.append((leaf_module, hook_handle))

    def save_out_analyse_log(self, save_log_path: str ="Analyse_layer_output.log"):
        out_flie = open(save_log_path, 'w')
        out_flie.write(str(tb_all))
        out_flie.close()



import hashlib
import math
import os
from logging import PlaceHolder

import linger
import linger.onnx
import onnx
import torch
import torch.nn
import torch.nn.functional as F
import torch.onnx
from linger.config import config
from linger.ops import (IQTensor, from_torch_tensor, iqadd, iqAddLayer, iqmul,
                        iqMulLayer)
from linger.utils import PlatFormQuant


def get_file_topolo_sort_type_list(f):
    model = onnx.load(f)
    return [n.op_type for n in model.graph.node]


def get_file_md5(fname):
    m = hashlib.md5()
    with open(fname, 'rb') as fobj:
        while True:
            data = fobj.read(4096)
            if not data:
                break
            m.update(data)
    return m.hexdigest()


if not os.path.exists("data.ignore"):
    os.mkdir("data.ignore")


def test_view_export():
    is_tuple = False

    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            if is_tuple:
                return args[0].view((1, 4, -1))
            else:
                return args[0].view(1, 4, -1)
    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 4)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_view.onnx", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #original_model = onnx.load("data.ignore/torch_view.onnx.ignore")
    aa_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_view.onnx")
    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_view.onnx", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq1_model = onnx.load("data.ignore/iq_view.onnx.ignore")
    aa_iq_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_view.onnx")
    assert aa_model_md5 == aa_iq_model_md5

    is_tuple = True
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_view_tuple.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)

    aa_iq_tuple_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_view_tuple.onnx.ignore")
    assert aa_model_md5 == aa_iq_tuple_model_md5


def test_view_as_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            return args[0].view_as(args[1])

    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 4)
    aa_copy = dummy_input.detach().data

    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, (aa_iq, aa_copy), "data.ignore/iq_view_as.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', 'y'], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    iq_model = onnx.load("data.ignore/iq_view_as.onnx.ignore")
    assert iq_model.graph.node[1].op_type == 'Reshape'


def test_relu_export():
    inplace = False

    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            if inplace:
                return torch.relu(*args)
            else:
                return torch.relu_(*args)
    net = TestModel()
    dummy_input = torch.randn(9, 8, 7, 6)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_relu.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #original_model = onnx.load("data.ignore/torch_relu.onnx.ignore")
    base_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_relu.onnx.ignore")
    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/torch_relu_iq1.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq1_model = onnx.load("data.ignore/torch_relu_iq1.onnx.ignore")
    iq1_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_relu_iq1.onnx.ignore")
    assert iq1_md5 == base_md5
    inplace = True
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/torch_relu_iq2.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq2_model = onnx.load("data.ignore/torch_relu_iq2.onnx.ignore")
    iq2_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_relu_iq2.onnx.ignore")
    assert iq2_md5 == base_md5


def test_maxpool_2d_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            return torch.max_pool2d(args[0], kernel_size=(2, 2), stride=2, padding=0)
    net = TestModel()
    dummy_input = torch.randn(1, 8, 16)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_maxpool_2d.onnx.ignore", export_params=True, opset_version=9,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    mode_base = onnx.load("data.ignore/torch_maxpool_2d.onnx.ignore")

    iq_1 = from_torch_tensor(dummy_input, 11, 8)
    with torch.no_grad():
        linger.onnx.export(net, iq_1, "data.ignore/qi1_maxpool_2d.onnx.ignore", export_params=True, opset_version=9,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    mode_iq = onnx.load("data.ignore/qi1_maxpool_2d.onnx.ignore")
    assert mode_iq.graph.node[0].op_type == mode_base.graph.node[0].op_type
    assert mode_iq.graph.node[0].attribute == mode_base.graph.node[0].attribute


def test_iq_mul_export_onnx():
    x = torch.tensor([[6.0, 4.0]], requires_grad=True).cuda()
    y = torch.tensor([[3.0, 8.0]], requires_grad=True).cuda()
    iq_layer = iqMulLayer()
    a0 = from_torch_tensor(x-1, 127.0/(6.0-1), 8)
    b0 = from_torch_tensor(y-1, 127.0/(8-1), 8)
    oscale = 127.0/(12-2)
    iq_layer(a0, b0, oscale)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)
    scale = 127.0 / 12
    with torch.no_grad():
        linger.onnx.export(iq_layer, (a, b, scale), "data.ignore/iq_mul1.onnx", export_params=True, keep_initializers_as_inputs=False,
                           opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    m_float = onnx.load("data.ignore/iq_mul1.onnx")
    assert m_float.graph.node[0].op_type == 'iqMul'
    assert len(m_float.graph.node[0].attribute) == 4
    for m in m_float.graph.node[0].attribute:
        if m.name == "scale_o":
            assert m.f == 128
        if m.name == "scale_x":
            assert abs(m.f - 127.0/6) < 0.01
        if m.name == "scale_y":
            assert abs(m.f - 127.0/8) < 0.01


def test_iq_mul_module_export_onnx():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, x, y):
            return iqmul(self, x, y, 'testname')
    x = torch.tensor([[6.0, 4.0]], requires_grad=True).cuda()
    y = torch.tensor([[3.0, 8.0]], requires_grad=True).cuda()
    net = TestModel()
    a0 = from_torch_tensor(x-1, 127.0/(6.0-1), 8)
    b0 = from_torch_tensor(y-1, 127.0/(8-1), 8)

    net(a0, b0)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)

    with torch.no_grad():
        linger.onnx.export(net, (a, b), "data.ignore/iq_mul2.onnx", export_params=True, keep_initializers_as_inputs=False,
                           opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    m_float = onnx.load("data.ignore/iq_mul2.onnx")
    assert m_float.graph.node[0].op_type == 'iqMul'
    assert len(m_float.graph.node[0].attribute) == 4
    assert len(m_float.graph.node) == 2
    for m in m_float.graph.node[0].attribute:
        if m.name == "scale_o":
            max_value = round(math.log(127/2.1, 2))
            scale_local = math.pow(2, max_value)
            assert abs(m.f - scale_local) < 0.1
        if m.name == "scale_x":
            assert abs(m.f - 127.0/6) < 0.01
        if m.name == "scale_y":
            assert abs(m.f - 127.0/8) < 0.01


def test_iq_mul_u8i8_module_export_onnx():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, x, y):
            return iqmul(self, x, y, 'testname')
    x = torch.tensor([[6.0, 4.0]], requires_grad=True).cuda()
    y = torch.tensor([[3.0, 8.0]], requires_grad=True).cuda()
    net = TestModel()
    a0 = from_torch_tensor(x-1, 127.0/(6.0-1), 8, 128)
    b0 = from_torch_tensor(y-1, 127.0/(8-1), 8)

    net(a0, b0)
    a = from_torch_tensor(x, 127.0/6.0, 8, 128)
    b = from_torch_tensor(y, 127.0/8, 8)

    with torch.no_grad():
        linger.onnx.export(net, (a, b), "data.ignore/iq_mul3.onnx", export_params=True, keep_initializers_as_inputs=False,
                           opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    m_float = onnx.load("data.ignore/iq_mul3.onnx")
    assert m_float.graph.node[0].op_type == 'iqMul'


def test_iq_add_export_onnx():
    linger.SetPlatFormQuant(platform_quant=PlatFormQuant.luna_quant)
    x = torch.tensor([[6.0, 4.0]], requires_grad=True).cuda()
    y = torch.tensor([[3.0, 8.0]], requires_grad=True).cuda()
    iq_layer = iqAddLayer()
    a0 = from_torch_tensor(x-1, 127.0/(6.0-1), 8)
    b0 = from_torch_tensor(y-1, 127.0/(8-1), 8)
    oscale = 127.0/(12-2)
    iq_layer(a0, b0, oscale)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)
    scale = torch.tensor([127.0/12])
    with torch.no_grad():
        linger.onnx.export(iq_layer, (a, b, scale), "data.ignore/iq_add1.onnx.ignore", export_params=True,
                           keep_initializers_as_inputs=False, opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    m_float = onnx.load("data.ignore/iq_add1.onnx.ignore")
    assert m_float.graph.node[0].op_type == 'iqAdd'
    assert len(m_float.graph.node[0].attribute) == 5
    for m in m_float.graph.node[0].attribute:
        if m.name == "scale_o":
            assert m.f == 128
        if m.name == "scale_x":
            assert abs(m.f - 127.0/6) < 0.01
        if m.name == "scale_y":
            assert abs(m.f - 127.0/8) < 0.01


def test_iq_add_module_export_onnx():
    linger.SetPlatFormQuant(platform_quant=PlatFormQuant.luna_quant)

    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, x, y):
            return iqadd(self, x, y, 'testname')
    x = torch.tensor([[6.0, 4.0]], requires_grad=True).cuda()
    y = torch.tensor([[3.0, 8.0]], requires_grad=True).cuda()
    net = TestModel()
    a0 = from_torch_tensor(x-1, 127.0/(6.0-1), 8)
    b0 = from_torch_tensor(y-1, 127.0/(8-1), 8)
    oscale = 127.0/(12-2)
    net(a0, b0)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)
    scale = torch.tensor([127.0/12])
    with torch.no_grad():
        linger.onnx.export(net, (a, b), "data.ignore/iq_add2.onnx.ignore", export_params=True, keep_initializers_as_inputs=False,
                           opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    m_float = onnx.load("data.ignore/iq_add2.onnx.ignore")
    assert m_float.graph.node[0].op_type == 'iqAdd'
    assert len(m_float.graph.node[0].attribute) == 5
    assert len(m_float.graph.node) == 2
    for m in m_float.graph.node[0].attribute:
        if m.name == "scale_o":
            assert m.f == 128.0
        if m.name == "scale_x":
            assert abs(m.f - 127.0/6) < 0.01
        if m.name == "scale_y":
            assert abs(m.f - 127.0/8) < 0.01


def test_reshape_export():
    is_tuple = False

    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            if is_tuple:
                return args[0].reshape((1, 4, -1))
            else:
                return args[0].reshape(1, 4, -1)
    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 4)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_reshape.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #original_model = onnx.load("data.ignore/torch_view.onnx.ignore")
    aa_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_reshape.onnx.ignore")
    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_reshape.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq1_model = onnx.load("data.ignore/iq_view.onnx.ignore")
    aa_iq_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_reshape.onnx.ignore")
    assert aa_model_md5 == aa_iq_model_md5

    is_tuple = True
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_reshape_tuple.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)

    aa_iq_tuple_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_reshape_tuple.onnx.ignore")
    assert aa_model_md5 == aa_iq_tuple_model_md5


def test_reshape_as_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            return args[0].reshape_as(args[1])

    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 4)
    aa_copy = dummy_input.detach().data

    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, (aa_iq, aa_copy), "data.ignore/iq_reshape_as.onnx.ignore", export_params=True,
                           opset_version=11, input_names=['x', 'y'], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    iq_model = onnx.load("data.ignore/iq_reshape_as.onnx.ignore")
    assert iq_model.graph.node[1].op_type == 'Reshape'


def test_squeeze_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            return args[0].squeeze()
    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 4)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_squeeze.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #original_model = onnx.load("data.ignore/torch_view.onnx.ignore")
    aa_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_squeeze.onnx.ignore")
    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_squeeze.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq1_model = onnx.load("data.ignore/iq_view.onnx.ignore")
    aa_iq_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_squeeze.onnx.ignore")
    assert aa_model_md5 == aa_iq_model_md5


def test_unsqueeze_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            return args[0].unsqueeze(dim=3)
    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 5)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_unsqueeze.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #original_model = onnx.load("data.ignore/torch_view.onnx.ignore")
    aa_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_unsqueeze.onnx.ignore")
    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_unsqueeze.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq1_model = onnx.load("data.ignore/iq_view.onnx.ignore")
    aa_iq_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_unsqueeze.onnx.ignore")
    assert aa_model_md5 == aa_iq_model_md5


def test_transpose_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            x = args[0].squeeze()
            return x.transpose(1, 2)
    net = TestModel()
    dummy_input = torch.randn(1, 2, 2, 2)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_transpose.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #original_model = onnx.load("data.ignore/torch_view.onnx.ignore")
    aa_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_transpose.onnx.ignore")
    aa_iq = from_torch_tensor(dummy_input, 5, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_transpose.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq1_model = onnx.load("data.ignore/iq_view.onnx.ignore")
    aa_iq_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_transpose.onnx.ignore")
    assert aa_model_md5 == aa_iq_model_md5


def test_flatten_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            return args[0].flatten(2, 4)
    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 7, 5)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_flatten.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #original_model = onnx.load("data.ignore/torch_view.onnx.ignore")
    aa_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/torch_flatten.onnx.ignore")
    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_flatten.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    #iq1_model = onnx.load("data.ignore/iq_view.onnx.ignore")
    aa_iq_model_md5 = get_file_topolo_sort_type_list(
        "data.ignore/iq_flatten.onnx.ignore")
    assert aa_model_md5 == aa_iq_model_md5


def test_getitem_export():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()

        def forward(self, *args):
            return args[0][:, 1, :, 2, :]
    net = TestModel()
    dummy_input = torch.randn(1, 2, 3, 7, 5)
    with torch.no_grad():
        linger.onnx.export(net, dummy_input, "data.ignore/torch_getitem.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    original_model = onnx.load("data.ignore/torch_getitem.onnx.ignore")

    aa_iq = from_torch_tensor(dummy_input, 12, 3)
    with torch.no_grad():
        linger.onnx.export(net, aa_iq, "data.ignore/iq_getitem.onnx.ignore", export_params=True, opset_version=11,
                           input_names=['x', ], operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    iq_model = onnx.load("data.ignore/iq_getitem.onnx.ignore")
    assert original_model.graph.node[0].op_type == iq_model.graph.node[0].op_type
    assert original_model.graph.node[1].op_type == iq_model.graph.node[1].op_type
    assert original_model.graph.node[2].op_type == iq_model.graph.node[2].op_type
    assert original_model.graph.node[3].op_type == iq_model.graph.node[3].op_type



from collections import OrderedDict

import torch
import torch.onnx
from torch.onnx import is_in_onnx_export

from ..config import config
from ..quant import Quant
from ..utils import *
from ..utils import Dump
from .iqtensor import from_torch_tensor, platform_to_string, quantlinear
from .ops import ModuleIntConfig


class ScaledRoundLayerFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, momentum, running_x, eval_scale_x, data_bits, training, prefix, dump, path, mode, quant, zero_point):
        if training:
            q_input, scale_x, max_value_x = quant.quant(
                input, data_bits, mode=mode, quant_data='input', iq_zero_point=zero_point)
            running_x.mul_(1-momentum).add_(momentum*max_value_x)
            scale_x = ScalerBuffer(scale_x)
            outputs = quant.dequant(q_input, scale_x)
        else:
            assert running_x != 0, 'invalid running_x=0, please finetune training before eval'
            scale_x = ScalerBuffer(quant.running_to_scale(
                running_x, data_bits, mode=mode, zero_point=zero_point))
            q_input, _, _ = quant.quant(
                input, data_bits, scale_x, mode=mode, quant_data='input', iq_zero_point=zero_point)
            outputs = quant.dequant(q_input, scale_x)

            if dump:
                name_list = ['input', 'q_input',
                             'outputs', 'scale_x', 'running_x']
                attr_list = [input, q_input, outputs, scale_x(), running_x()]
                Dump.dump_file(prefix, "ScaledRoundLayer.",
                               zip(name_list, attr_list), path)
            eval_scale_x.fill_(scale_x())
        if isinstance(scale_x, float):
            return from_torch_tensor(outputs, scale_x, data_bits, zero_point)
        elif isinstance(scale_x, torch.Tensor):
            return from_torch_tensor(outputs, scale_x.item(), data_bits, zero_point)
        else:
            return from_torch_tensor(outputs, scale_x.data, data_bits, zero_point)

    @staticmethod
    def backward(ctx, grad_output):
        return (grad_output, None, None, None, None, None, None, None, None, None, None, None)

    @staticmethod
    def symbolic(g, input, momentum, running_x, scale_x, data_bits, training, prefix, dump, path, mode, quant, zero_point):
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        return quantlinear(g, input, scale_x(), platform_quant, data_bits)


class ScaledRoundLayer(torch.nn.Module):
    r"""
        将浮点的tensor依据统计的scale转为定点数值

    Args:
        bits: 输出量化的位数
        mode：量化方式，支持MaxValue和Qvalue
    """

    def __init__(self, bits=8, mode=QuantMode.QValue, zero_point=0):
        super(ScaledRoundLayer, self).__init__()

        self.prefix = ""
        self.dump = False
        self.path = ""
        self.data_bits = bits
        self.quant_mode = mode
        self.momentum = 0.1
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('running_x', torch.zeros(1))
        self.quant = Quant()
        self.zero_point = zero_point

    def forward(self, x, *others):
        assert self.data_bits == 8, 'quant_tensor only support 8bit'
        scale_x = ScalerBuffer(self.scale_x)
        x = ScaledRoundLayerFunction.apply(x, self.momentum, self.running_x, scale_x, self.data_bits, self.training,
                                           self.prefix, self.dump, self.path, self.quant_mode, self.quant, self.zero_point)
        self.scale_x.fill_(scale_x.data)
        if len(others) == 0:
            return x
        else:
            return tuple([x])+others

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=self._version)
        if is_in_onnx_export():
            assert self._buffers['running_x'] > 0, 'invalid running_x, please finetune first'
            scale_x = ScalerBuffer(self.quant.running_to_scale(ScalerBuffer(
                self._buffers['running_x']), self.data_bits, mode=self.quant_mode, zero_point=self.zero_point))
            self._buffers['scale_x'].data.fill_(scale_x())

        self._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in self._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in self._state_dict_hooks.values():
            hook_result = hook(self, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


__all__ = ['ScaledRoundLayer']



import os

import linger
import numpy as np
import torch
import torch.nn as nn
from linger.ops import IQTensor, from_torch_tensor

if not os.path.exists('data.ignore'):
    os.mkdir('data.ignore')


def test_iqadd_load_state_dict_1():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.transpose = nn.ConvTranspose2d(2, 2, 5, 5, 2, 4, 2, True, 2)
            self.conv = nn.Conv2d(2, 2, kernel_size=3, stride=1,
                                  padding=1, bias=True)
            self.fc = nn.Linear(392, 100, bias=False)

        def forward(self, x):
            trans = self.transpose(x)
            assert isinstance(trans, IQTensor)
            conv = self.conv(trans)
            assert isinstance(conv, IQTensor)
            x = trans + conv
            assert isinstance(x, IQTensor)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))
            x = self.fc(x)
            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    torch.cuda.set_device(0)
    net1 = Net().cuda()
    dummy_input = torch.randn(1, 2, 2, 2).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()

    replace_tuple = (nn.Linear, nn.ConvTranspose2d, nn.Conv2d)
    net1.train()
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    net1 = linger.init(net1, quant_modules=replace_tuple,
                       mode=linger.QuantMode.QValue)
    optimizer1 = torch.optim.SGD(net1.parameters(), lr=0.01)
    for i in range(150):
        optimizer1.zero_grad()
        out1 = net1(dummy_input)
        loss1 = criterion(out1, target)
        loss1.backward()
        optimizer1.step()
        if i % 30 == 29:
            print('loss1 {}'.format(loss1))
    net1.eval()
    torch.save(net1.state_dict(), 'data.ignore/model.pt.ignore')
    out1 = net1(dummy_input)

    net2 = Net().cuda()
    net2 = linger.init(net2, quant_modules=replace_tuple,
                       mode=linger.QuantMode.QValue)
    net2.load_state_dict(torch.load(
        'data.ignore/model.pt.ignore', map_location='cpu'))
    net2.eval()
    out2 = net2(dummy_input)

    with torch.no_grad():
        torch.onnx.export(net2, dummy_input, "data.ignore/iqadd_t710.onnx", export_params=True,
                          opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)

    assert out1.sum() == out2.sum()


LINGER = '_linger'

LINGER_MIX_INT8_MANUAL_ROUND_LAYERS = LINGER+'_round_tensor'

LINGER_DUMP_NAME = LINGER+"_dump_name"

LINGER_IGNORE_PAMAMTER = LINGER+"_ignore_parameter"

LINGER_IQTENSOR_LAYER_COUNTER = LINGER+"_iq_tensor_index_"

LINGER_FUNCINT_BMM_COUNTER = LINGER+"_funcint_bmm_index_"

LINGER_MODE = LINGER+"_mode"

LINGER_OBIT = LINGER+"_obit"

LINGER_AHEAD_RELU = LINGER+"_ahead_relu"

LINGER_AHEAD_SIGMOID = LINGER+"_ahead_sigmoid"


__all__ = ['LINGER_MIX_INT8_MANUAL_ROUND_LAYERS', 'LINGER_DUMP_NAME', 'LINGER_IGNORE_PAMAMTER', 'LINGER_IQTENSOR_LAYER_COUNTER',
           'LINGER_FUNCINT_BMM_COUNTER', 'LINGER_OBIT', 'LINGER_MODE', 'LINGER_AHEAD_RELU', 'LINGER_AHEAD_SIGMOID']



import  onnx
from onnx import numpy_helper
import numpy as np
from onnx import TensorProto
from .onnx_utils import get_node_id,remove_identity_node


def get_quant_linear_node(onnx_model, node, scale_x, scale_w, scale_o, platform_quant):
    # scale_x = 16.0
    # scale_w = 64.0
    # scale_o = 16.0
    domain = "thinker"
    activation_bits = 8
    weight_bits = 8

    # get weights of conv
    weight_data = None
    bias_data = None
    
    # for init in reversed(onnx_model.graph.initializer):
    for init in onnx_model.graph.initializer[::-1]:

        if init.name == node.input[1]:
            weight_data = numpy_helper.to_array(init)
            onnx_model.graph.initializer.remove(init)
        try:
            if init.name == node.input[2]:
                bias_data = numpy_helper.to_array(init)
                onnx_model.graph.initializer.remove(init)
        except:
            pass

    # initializer quantization
    # the definition of scale is different from the normal !!!
    # todo, how to handle data type
    if weight_data.dtype == "float32":
        weight_data = np.floor(weight_data * scale_w + 0.5).astype(np.int8)    
    weight_tensor = onnx.helper.make_tensor(node.input[1], TensorProto.INT8, weight_data.shape, \
        weight_data.tobytes(), raw=True)  # raw 
    try:
        if bias_data.dtype == "float32":
            bias_data = np.floor(bias_data * scale_w * scale_x + 0.5).astype(np.int32)
        bias_tensor = onnx.helper.make_tensor(node.input[2], TensorProto.INT32, bias_data.shape, bias_data.tobytes(), raw=True)     
    except:
        pass           
    if node.op_type == "Gemm":
        if len(node.input) == 3:
            quant_node = onnx.helper.make_node(
                "LinearInt",
                name=node.name,
                inputs=[node.input[0], node.input[1], node.input[2]],
                outputs=[node.output[0]],
                domain=domain
            )
        else:
            quant_node = onnx.helper.make_node(
                "LinearInt",
                name=node.name,
                inputs=[node.input[0], node.input[1]],
                outputs=[node.output[0]],
                domain=domain
            )

    # add attributes of original node to the quant node
    quant_node.attribute.extend(
        onnx.helper.make_attribute(attr.name, onnx.helper.get_attribute_value(attr)) for attr in node.attribute
    )
    # add attributes that original node does not have to the quant node
    quant_node.attribute.extend(
        [onnx.helper.make_attribute("scale_x", scale_x), onnx.helper.make_attribute("scale_w", scale_w), onnx.helper.make_attribute("scale_o", scale_o), \
            onnx.helper.make_attribute("data_bits", activation_bits), onnx.helper.make_attribute("o_bits", activation_bits), onnx.helper.make_attribute("parameter_bits", weight_bits), \
            onnx.helper.make_attribute("platform_quant", platform_quant)]
    )

    # insert the quant initializers
    try:
        onnx_model.graph.initializer.extend([weight_tensor, bias_tensor])
    except:
        onnx_model.graph.initializer.extend([weight_tensor])
    return quant_node

def insert_op_before(model, target_node_index, ori_node, scale_x, scale_w, scale_o, platform_quant ):
    '''
    op_name
    weight_dict
    attr_dict
    '''

    replace_node = get_quant_linear_node(model, ori_node, scale_x=scale_x, scale_w=scale_w, scale_o=scale_o, platform_quant=platform_quant)
    
    model.graph.node.insert(target_node_index, replace_node)

def replace_linearint(module, scale_x, scale_w, scale_o, platform_quant):
    nodes  = module.graph.node[::-1]
    remove_node_all=[]
    for i,node in enumerate(nodes):
        if node.op_type == 'Gemm' :   # or node.op_type == "Matmul":
            remove_node_all.append((i,node))

    from onnx import numpy_helper
    input_tensors = {  t.name: numpy_helper.to_array(t) for t in module.graph.initializer }
            
    for i_node in remove_node_all:
        i    = i_node[0]
        node = i_node[1] 

        origin_layer_index = get_node_id(node,module.graph.node)
        module.graph.node.remove(node)        
        insert_op_before(
                    module,
                    target_node_index = origin_layer_index,
                    ori_node = node,
                    scale_x = scale_x, 
                    scale_w = scale_w, 
                    scale_o = scale_o, 
                    platform_quant = platform_quant
                     )
   
    return module


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from typing import Tuple

import torch
import torch.nn as nn

from .config import config
from .modules import *
from .ops.ops_names import LINGER_AHEAD_RELU, LINGER_AHEAD_SIGMOID
from .utils import ClampInfo, Singleton, get_device, logger


class _SingletonContainClampModules(Singleton):
    clamped_quant_list = {}
    _is_close_register = False

    def _close_register(self):
        self._is_close_register = True

    def _register(self, module, quant_info):
        assert isinstance(module, torch.nn.Module) or isinstance(module, list)
        modules = [module] if isinstance(module, torch.nn.Module) else module
        for each_mod in modules:
            if self._is_close_register:
                print("warning: module has initlized and linger.init may not work")
            self.clamped_quant_list[each_mod] = quant_info

    def _is_registered(self, module):
        return (module in self.clamped_quant_list.keys())

    def get(self, module):
        return self.clamped_quant_list.get(module)

    def clear(self):
        if self._is_close_register:
            print("warning: module has initlized and linger.clear may not work")
        self.clamped_quant_list.clear()


def disable_normalize(module: nn.Module):
    r"""
        禁用clamp策略替换网络，该接口支持inline 使用
    Notes:
        disable_normalize 应该在linger.normalize_layers函数之前调用才生效
    """
    queue = [module]
    while len(queue) > 0:
        node = queue.pop(0)
        if type(node) in SupportNormalizeTorchModules:
            _SingletonContainClampModules()._register(node, None)
        for _, submodule in node.named_children():
            queue.append(submodule)


def normalize_module(module: nn.Module, type_modules: Tuple = DefaultNormalizeIntXModule, normalize_weight_value: float = 8, normalize_bias_value: float = 8, normalize_output_value: float = None):
    r"""
        对module进行自定义clamp设置，通用的clamp策略接口，该接口支持inline使用
    """
    if not isinstance(type_modules, tuple):
        type_modules = (type_modules, )
    for t in type_modules:
        if t not in SupportNormalizeTorchModules:
            logger.fatal(str(t) + 'is not support clamp in linger now')
            exit(-1)
    queue = [module]
    while len(queue) > 0:
        node = queue.pop(0)
        if type(node) in type_modules and type(node) in SupportNormalizeTorchModules:
            cinfo = ClampInfo()
            cinfo.set_clamp_weight_value(normalize_weight_value)
            cinfo.set_clamp_bias_value(normalize_bias_value)
            cinfo.set_clamp_output_value(normalize_output_value)
            _SingletonContainClampModules()._register(node, cinfo)
        for _, submodule in node.named_children():
            queue.append(submodule)


def _replaceModule(submodule, normalize_weight_value, normalize_bias_value, normalize_output_value):
    if isinstance(submodule, tuple(SupportNormalizeTorchModules)):
        if isinstance(submodule, NormalizeConvBN2d):
            bias = True if submodule.conv.bias is not None else False
            if config.BnMomentumUpdate.disable:
                submodule_bn_momentum = 0
            else:
                submodule_bn_momentum = submodule.bn.momentum
            convbn = NormalizeConvBN2d(submodule.conv.in_channels, submodule.conv.out_channels, submodule.conv.kernel_size, submodule.conv.stride, submodule.conv.padding, submodule.conv.dilation,
                                       submodule.conv.groups, bias, submodule.conv.padding_mode, submodule.bn.eps, submodule_bn_momentum, submodule.bn.affine, submodule.bn.track_running_stats,
                                       normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value, ahead_relu=submodule.ahead_relu)
            return convbn
        elif isinstance(submodule, NormalizeConvBN1d):
            bias = True if submodule.conv.bias is not None else False
            if config.BnMomentumUpdate.disable:
                submodule_bn_momentum = 0
            else:
                submodule_bn_momentum = submodule.bn.momentum
            convbn = NormalizeConvBN1d(submodule.conv.in_channels, submodule.conv.out_channels, submodule.conv.kernel_size, submodule.conv.stride, submodule.conv.padding, submodule.conv.dilation,
                                       submodule.conv.groups, bias, submodule.conv.padding_mode, submodule.bn.eps, submodule_bn_momentum, submodule.bn.affine, submodule.bn.track_running_stats,
                                       normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value, ahead_relu=submodule.ahead_relu)
            return convbn
        elif isinstance(submodule, nn.Conv2d):
            bias = True if submodule.bias is not None else False
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            ahead_sigmoid = getattr(
                submodule, LINGER_AHEAD_SIGMOID, False)

            conv = NormalizeConv2d(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.dilation,
                                   submodule.groups, bias, submodule.padding_mode, normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value, ahead_relu=ahead_relu, ahead_sigmoid=ahead_sigmoid)
            return conv
        elif isinstance(submodule, nn.Linear):
            bias = True if submodule.bias is not None else False
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            ahead_sigmoid = getattr(
                submodule, LINGER_AHEAD_SIGMOID, False)

            linear = NormalizeLinear(submodule.in_features, submodule.out_features, bias, normalize_data=normalize_output_value, normalize_weight=normalize_weight_value,
                                     normalize_bias=normalize_bias_value, ahead_relu=ahead_relu, ahead_sigmoid=ahead_sigmoid)
            return linear
        elif isinstance(submodule, nn.Embedding):
            embed = NormalizeEmbedding(submodule.num_embeddings, submodule.embedding_dim, submodule.padding_idx, submodule.max_norm, submodule.norm_type, submodule.scale_grad_by_freq,
                                       submodule.sparse, submodule.weight, normalize_data=normalize_output_value, normalize_weight=normalize_weight_value)
            return embed
        elif isinstance(submodule, nn.ConvTranspose2d):
            bias = True if submodule.bias is not None else False
            convtranspose = NormalizeConvTranspose2d(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding,
                                                     submodule.output_padding, submodule.groups, bias, submodule.dilation, submodule.padding_mode, normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value)
            return convtranspose
        elif isinstance(submodule, nn.BatchNorm2d):
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            if config.BnMomentumUpdate.disable:
                submodule_momentum = 0
            else:
                submodule_momentum = submodule.momentum
            bn = NormalizeBatchNorm2d(submodule.num_features, eps=submodule.eps, momentum=submodule_momentum, affine=submodule.affine, track_running_stats=submodule.track_running_stats,
                                      normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value, ahead_relu=ahead_relu)
            return bn

        elif isinstance(submodule, nn.Conv1d):
            bias = True if submodule.bias is not None else False
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            conv = NormalizeConv1d(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.dilation,
                                   submodule.groups, bias, submodule.padding_mode, normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value, ahead_relu=ahead_relu)
            return conv
        elif isinstance(submodule, nn.GRU):
            gru = NormalizeFastGRU(submodule.input_size, submodule.hidden_size, submodule.num_layers, submodule.bias, submodule.batch_first, submodule.dropout, submodule.bidirectional,
                                   normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value)
            return gru
        elif isinstance(submodule, nn.LSTM):
            lstm = NormalizeFastLSTM(submodule.input_size, submodule.hidden_size, submodule.num_layers, submodule.bias, submodule.batch_first, submodule.dropout, submodule.bidirectional,
                                     normalize_data=normalize_output_value, normalize_weight=normalize_weight_value, normalize_bias=normalize_bias_value)
            return lstm

    return None


def normalize_layers(model: nn.Module, *, normalize_modules: Tuple = DefaultNormalizeIntXModule, normalize_weight_value: float = 8, normalize_bias_value: float = None, normalize_output_value: float = 8) -> nn.Module:
    r"""对模型进行clamp 处理以方便进行更好的量化，
        默认支持nn.Conv2d, nn.Linear, nn.ConvTranspose2d, Conv-BN融合Clamp

    """
    if type(normalize_modules) is not tuple:
        normalize_modules = (normalize_modules, )
    normalize_modules = set(list(normalize_modules) +
                            [NormalizeConvBN2d, NormalizeConvBN1d])
    for user_module_type in normalize_modules:
        assert user_module_type in SupportNormalizeTorchModules, 'Currently not support clamp of ' + \
            str(user_module_type)
    device = get_device(model)
    queue = [model]
    while len(queue) > 0:
        node = queue.pop(0)
        for name, submodule in node.named_children():
            if _SingletonContainClampModules()._is_registered(submodule):
                clamp_info = _SingletonContainClampModules().get(submodule)
                if clamp_info == None:
                    continue
                else:
                    r_module = _replaceModule(submodule, clamp_info.clamp_weight_value, clamp_info.clamp_bias_value,
                                              clamp_info.clamp_output_value)
                    assert r_module is not None
                    setattr(node, name, r_module)
            elif type(submodule) in SupportNormalizeTorchModules and type(submodule) in normalize_modules:
                r_module = _replaceModule(
                    submodule, normalize_weight_value, normalize_bias_value, normalize_output_value)
                assert r_module is not None
                setattr(node, name, r_module)
            else:
                queue.append(submodule)

    if model.training:
        model.train()
    else:
        model.eval()
    model.to(device)
    return model



import onnx
from onnx import numpy_helper

def replace_clip_attr(model,scale_o): #only support opset 9 and below
    if model.opset_import[0].version <12:
        assert False, "Clip quant only support opset 12 and above !"
    
    for index,node in enumerate(model.graph.node[::-1]):
        if node.op_type == "Clip":
            model.graph.node[::-1][index+1].attribute[0].i = 3
            model.graph.node[::-1][index+2].attribute[0].i = 3
            # max_value = int.from_bytes(model.graph.node[::-1][index+3].attribute[0].t.raw_data, byteorder='little', signed=True)
            # min_value = int.from_bytes(model.graph.node[::-1][index+4].attribute[0].t.raw_data, byteorder='little', signed=True)

            max_value = numpy_helper.to_array(model.graph.node[::-1][index+3].attribute[0].t)
            min_value = numpy_helper.to_array(model.graph.node[::-1][index+4].attribute[0].t)

            max_value = max_value * scale_o
            min_value = min_value * scale_o
            
            model.graph.node[::-1][index+3].attribute[0].t.data_type = 11  #double
            model.graph.node[::-1][index+4].attribute[0].t.data_type = 11

            model.graph.node[::-1][index+3].attribute[0].t.raw_data = max_value.tobytes()
            model.graph.node[::-1][index+4].attribute[0].t.raw_data= min_value.tobytes()

    return model


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import linger
import numpy as np
import torch
import torch.nn as nn


def test_clip():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.transpose = nn.ConvTranspose2d(2, 2, 5, 5, 2, 4, 2, True, 2)
            self.conv = nn.Conv2d(2, 2, kernel_size=3, stride=1,
                                  padding=1, bias=True)
            self.bn = nn.BatchNorm2d(2)
            self.relu = nn.ReLU()
            self.fc = nn.Linear(392, 100, bias=True)

        def forward(self, x):
            x = self.transpose(x)
            x = self.conv(x)
            x = self.bn(x)
            x = self.relu(x)
            n, c, h, w = x.shape
            x = x.view(n, c*h*w)
            x = self.fc(x)
            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    # random.seed(1)

    torch.cuda.set_device(0)
    net2 = Net().cuda()
    dummy_input = torch.randn(1, 2, 2, 2).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.ConvTranspose2d, nn.Linear)
    net2.train()
    net2 = linger.init(net2, quant_modules=replace_tuple, mode=linger.QuantMode.QValue)

    optimizer2 = torch.optim.SGD(net2.parameters(), lr=0.01)
    for i in range(150):
        optimizer2.zero_grad()
        out2 = net2(dummy_input)
        loss2 = criterion(out2, target)
        loss2.backward()
        optimizer2.step()
        if i % 30 == 29:
            print('loss2 {}'.format(loss2))

    net2.eval()
    out2 = net2(dummy_input)
    assert abs(out2.mean() - target.mean()) < 0.01
    # reset avoid other test files params be changed


from .modules_configs import (DefaultNormalizeIntXModule,
                              SupportNormalizeIntXModules,
                              SupportNormalizeTorchModules)
from .normalize_batchnorm2d import NormalizeBatchNorm2d
from .normalize_conv1d import NormalizeConv1d
from .normalize_conv2d import NormalizeConv2d
from .normalize_convbn1d import NormalizeConvBN1d
from .normalize_convbn2d import NormalizeConvBN2d
from .normalize_convTranspose2d import NormalizeConvTranspose2d
from .normalize_embedding import NormalizeEmbedding
from .normalize_fastGRU import NormalizeFastGRU
from .normalize_fastLSTM import NormalizeFastLSTM
from .normalize_layernorm import NormalizeLayerNorm
from .normalize_linear import NormalizeLinear


import itertools
from typing import Tuple

import torch
import torch.nn as nn

import linger

from .config import config
from .modules import *
from .onnx import export as linger_export
from .ops import *
from .ops.iqtensor import iqAddLayer, iqDivLayer, iqMulLayer, iqSumLayer
from .ops.linger_functional import bmm as linger_bmm
from .ops.linger_functional import cat as linger_cat
from .ops.linger_functional import channel_shuffle_quant
from .ops.linger_functional import clamp as linger_clamp
from .ops.linger_functional import clamp_ as linger_clamp_
from .ops.linger_functional import dropout as linger_dropout
from .ops.linger_functional import (iqCatLayer, iqClampLayer, iqSigmoidLayer,
                                    iqTanhLayer, iqVarLayer)
from .ops.linger_functional import logsoftmax as linger_logsoftmax
from .ops.linger_functional import \
    pack_padded_sequence as linger_pack_padded_sequence
from .ops.linger_functional import \
    pad_packed_sequence as linger_pad_packed_sequence
from .ops.linger_functional import sigmoid as linger_sigmoid
from .ops.linger_functional import sigmoid_ as linger_sigmoid_
from .ops.linger_functional import softmax as linger_softmax
from .ops.linger_functional import tanh as linger_tanh
from .ops.linger_functional import tanh_ as linger_tanh_
from .ops.linger_functional import (torch_pack_padded_sequence,
                                    torch_pad_packed_sequence)
from .ops.linger_functional import var as linger_var
from .ops.module_self import hook_forward, hook_pre_forward
from .ops.ops_names import (LINGER_AHEAD_RELU, LINGER_AHEAD_SIGMOID,
                            LINGER_MIX_INT8_MANUAL_ROUND_LAYERS, LINGER_MODE,
                            LINGER_OBIT)
from .utils import QuantInfo, QuantMode, Singleton, get_device, logger

__all__ = ["disable_quant", "quant_module",
           "quant_module_by_type", "quant_tensor", "init"]


class _SingletonContainCustomModules(Singleton):
    customized_quant_list = {}
    _is_close_register = False

    def _close_register(self):
        self._is_close_register = True

    def _register(self, module, quant_info):
        assert isinstance(module, torch.nn.Module) or isinstance(module, list)
        modules = [module] if isinstance(module, torch.nn.Module) else module
        for each_mod in modules:
            if self._is_close_register:
                print("warning: module has initlized and linger.init may not work")
            self.customized_quant_list[each_mod] = quant_info

    def _is_registered(self, module):
        return (module in self.customized_quant_list.keys())

    def get(self, module):
        return self.customized_quant_list.get(module)

    def clear(self):
        if self._is_close_register:
            print("warning: module has initlized and linger.clear may not work")
        self.customized_quant_list.clear()


def fuse_state_dict(state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):

    eps = 1e-5
    clamp_conv_name = prefix + 'conv'
    clamp_bn_name = prefix + 'bn'
    conv_int_name = prefix
    if clamp_conv_name + '.weight' in state_dict and clamp_bn_name + '.weight' in state_dict:
        b_mean = state_dict[clamp_bn_name + '.running_mean']
        b_var = state_dict[clamp_bn_name + '.running_var']
        b_w = state_dict[clamp_bn_name + '.weight']
        b_b = state_dict[clamp_bn_name + '.bias']
        sigma = 1 / torch.sqrt(b_var + eps)
        alpha = b_w * sigma
        beta = b_b - b_mean * alpha
        c_w = state_dict[clamp_conv_name + '.weight']
        state_dict[conv_int_name +
                   'weight'] = (c_w * alpha.view(-1, *([1]*(len(c_w.shape)-1))))
        if clamp_conv_name + '.bias' in state_dict:
            c_b = state_dict[clamp_conv_name + '.bias']
            state_dict[conv_int_name + 'bias'] = (c_b * alpha + beta)
            state_dict.pop(clamp_conv_name + '.bias')
        else:
            state_dict[conv_int_name + 'bias'] = beta
        state_dict.pop(clamp_bn_name + '.running_mean')
        state_dict.pop(clamp_bn_name + '.running_var')
        state_dict.pop(clamp_bn_name + '.weight')
        state_dict.pop(clamp_bn_name + '.bias')
        state_dict.pop(clamp_bn_name + '.num_batches_tracked')
        state_dict.pop(clamp_conv_name + '.weight')
    else:
        assert clamp_conv_name + '.weight' not in state_dict and clamp_bn_name + \
            '.weight' not in state_dict, 'load quanted model but contain float clamp params'


def _replaceOp(submodule, mode, in_data_bits, parameter_bits, out_bits=None):
    assert in_data_bits > 0 and in_data_bits <= 32, "in_data_bits should between 0 and 32"
    assert parameter_bits > 0 and parameter_bits <= 32, "parameter_bits should between 0 and 32"
    assert out_bits is None or out_bits > 0 and out_bits <= 32, "out_bits should between 0 and 32"
    assert mode == QuantMode.QValue, "mode support only q_value"
    if isinstance(submodule, tuple(SupportQuantTorchModules)):
        if isinstance(submodule, NormalizeFastGRU):
            gru = GRUInt(submodule.input_size, submodule.hidden_size, submodule.num_layers, submodule.bias, submodule.batch_first, submodule.dropout, submodule.bidirectional,
                         data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                         clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias)
            return gru
        elif isinstance(submodule, NormalizeFastLSTM):
            lstm = LSTMInt(submodule.input_size, submodule.hidden_size, submodule.num_layers, submodule.bias, submodule.batch_first, submodule.dropout, submodule.bidirectional,
                           data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                           clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias)
            return lstm
        elif isinstance(submodule, NormalizeConvBN1d):
            # ahead_relu = getattr(submodule,LINGER_AHEAD_RELU,False)
            conv = Conv1dInt(submodule.conv.in_channels, submodule.conv.out_channels, submodule.conv.kernel_size, submodule.conv.stride, submodule.conv.padding, submodule.conv.dilation, submodule.conv.groups,
                             True, submodule.conv.padding, data_bits=in_data_bits, parameter_bits=parameter_bits, o_bits=out_bits, mode=mode,
                             clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias, ahead_relu=submodule.ahead_relu)
            conv._register_load_state_dict_pre_hook(fuse_state_dict)
            return conv
        elif isinstance(submodule, NormalizeConvBN2d):
            # ahead_relu = getattr(submodule,LINGER_AHEAD_RELU,False)
            conv = Conv2dInt(submodule.conv.in_channels, submodule.conv.out_channels, submodule.conv.kernel_size, submodule.conv.stride, submodule.conv.padding, submodule.conv.dilation, submodule.conv.groups,
                             True, submodule.conv.padding, data_bits=in_data_bits, parameter_bits=parameter_bits, o_bits=out_bits, mode=mode,
                             clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias, ahead_relu=submodule.ahead_relu)
            conv._register_load_state_dict_pre_hook(fuse_state_dict)
            return conv
        elif isinstance(submodule, NormalizeConv1d):
            bias = True if submodule.bias is not None else False
            # ahead_relu = getattr(submodule,LINGER_AHEAD_RELU,False)
            conv = Conv1dInt(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.dilation, submodule.groups,
                             bias, submodule.padding_mode, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                             clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias, ahead_relu=submodule.ahead_relu)
            return conv
        elif isinstance(submodule, NormalizeConv2d):
            bias = True if submodule.bias is not None else False
            # ahead_relu = getattr(submodule,LINGER_AHEAD_RELU,False)
            conv = Conv2dInt(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.dilation, submodule.groups,
                             bias, submodule.padding_mode, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                             clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias,
                             ahead_relu=submodule.ahead_relu, ahead_sigmoid=submodule.ahead_sigmoid)
            return conv
        elif isinstance(submodule, NormalizeConvTranspose2d):
            bias = True if submodule.bias is not None else False
            conv_transpose = ConvTranspose2dInt(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.output_padding, submodule.groups,
                                                bias, submodule.dilation, submodule.padding_mode, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                                                clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias)
            return conv_transpose
        elif isinstance(submodule, NormalizeLinear):
            bias = True if submodule.bias is not None else False
            linear = LinearInt(submodule.in_features, submodule.out_features, bias, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                               clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias,
                               ahead_relu=submodule.ahead_relu, ahead_sigmoid=submodule.ahead_sigmoid)
            return linear
        elif isinstance(submodule, NormalizeBatchNorm2d):
            # ahead_relu = getattr(submodule,LINGER_AHEAD_RELU,False)
            if config.BnMomentumUpdate.disable:
                submodule_momentum = 0
            else:
                submodule_momentum = submodule.momentum
            bn = BatchNormInt(submodule.num_features, eps=submodule.eps, momentum=submodule_momentum, affine=submodule.affine, track_running_stats=submodule.track_running_stats,
                              data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                              clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight, clamp_bias=submodule.normalize_bias, ahead_relu=submodule.ahead_relu)
            return bn
        elif isinstance(submodule, NormalizeLayerNorm):
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            submodule_momentum = 0.1
            layer_norm = LayerNormInt(submodule.normalized_shape, submodule.eps, submodule_momentum, submodule.elementwise_affine,
                                      data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits, ahead_relu=ahead_relu)
            return layer_norm
        elif isinstance(submodule, NormalizeEmbedding):
            embedding = EmbeddingInt(submodule.num_embeddings, submodule.embedding_dim, submodule.padding_idx, submodule.max_norm, submodule.norm_type, submodule.scale_grad_by_freq, submodule.sparse,
                                     submodule.weight, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                                     clamp_data=submodule.normalize_data, clamp_weight=submodule.normalize_weight)
            return embedding
        elif isinstance(submodule, nn.Linear):
            bias = True if submodule.bias is not None else False
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            ahead_sigmoid = getattr(
                submodule, LINGER_AHEAD_SIGMOID, False)
            linear = LinearInt(submodule.in_features, submodule.out_features, bias, data_bits=in_data_bits,
                               parameter_bits=parameter_bits, mode=mode, o_bits=out_bits, ahead_relu=ahead_relu, ahead_sigmoid=ahead_sigmoid)
            return linear
        elif isinstance(submodule, nn.Conv2d):
            bias = True if submodule.bias is not None else False
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            ahead_sigmoid = getattr(
                submodule, LINGER_AHEAD_SIGMOID, False)
            conv = Conv2dInt(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.dilation, submodule.groups,
                             bias, submodule.padding_mode, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits,
                             ahead_relu=ahead_relu, ahead_sigmoid=ahead_sigmoid)
            return conv
        elif isinstance(submodule, nn.BatchNorm2d):
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            if config.BnMomentumUpdate.disable:
                submodule_momentum = 0
            else:
                submodule_momentum = submodule.momentum
            batch_norm = BatchNormInt(submodule.num_features, submodule.eps, submodule_momentum, submodule.affine, submodule.track_running_stats,
                                      data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits, ahead_relu=ahead_relu)
            return batch_norm

        elif isinstance(submodule, nn.LayerNorm):
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            submodule_momentum = 0.1
            layer_norm = LayerNormInt(submodule.normalized_shape, submodule.eps, submodule_momentum, submodule.elementwise_affine,
                                      data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits, ahead_relu=ahead_relu)
            return layer_norm
        elif isinstance(submodule, nn.ConvTranspose2d):
            bias = True if submodule.bias is not None else False
            conv_transpose = ConvTranspose2dInt(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.output_padding, submodule.groups,
                                                bias, submodule.dilation, submodule.padding_mode, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits)
            return conv_transpose
        elif isinstance(submodule, nn.GRU):
            gru = GRUInt(submodule.input_size, submodule.hidden_size, submodule.num_layers, submodule.bias, submodule.batch_first,
                         submodule.dropout, submodule.bidirectional, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits)
            return gru
        elif isinstance(submodule, nn.LSTM):
            lstm = LSTMInt(submodule.input_size, submodule.hidden_size, submodule.num_layers, submodule.bias, submodule.batch_first,
                           submodule.dropout, submodule.bidirectional, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits)
            return lstm
        elif isinstance(submodule, nn.AvgPool2d):
            avg_pool = AvgPool2dInt(submodule.kernel_size, submodule.stride, submodule.padding, submodule.ceil_mode,
                                    submodule.count_include_pad, submodule.divisor_override, data_bits=in_data_bits, mode=mode, o_bits=out_bits)
            return avg_pool
        elif isinstance(submodule, nn.Conv1d):
            bias = True if submodule.bias is not None else False
            ahead_relu = getattr(submodule, LINGER_AHEAD_RELU, False)
            conv1d = Conv1dInt(submodule.in_channels, submodule.out_channels, submodule.kernel_size, submodule.stride, submodule.padding, submodule.dilation, submodule.groups,
                               bias, submodule.padding_mode, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits, ahead_relu=ahead_relu)
            return conv1d
        elif isinstance(submodule, nn.ReLU6):
            relu6 = ReLU6Int(data_bits=in_data_bits, mode=mode)
            return relu6
        elif isinstance(submodule, nn.Embedding):
            embedding = EmbeddingInt(submodule.num_embeddings, submodule.embedding_dim, submodule.padding_idx, submodule.max_norm, submodule.norm_type, submodule.scale_grad_by_freq, submodule.sparse,
                                     submodule.weight, data_bits=in_data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=out_bits)
            return embedding

    return None


def disable_quant(module: nn.Module):
    r"""禁用module及子module 任何量化策略(使用原有浮点精度),该接口支持inline使用(Example2)

    Args:
        module:需要不进行量化的module

    Example:        
        >>> net = shufflenet_v2_x1_0(pretrained=False)
        >>> linger.disable_quant(net)
    Example:
        >>> class Net(nn.Module):
        >>>     def __init__(self):
        >>>         super(Net, self).__init__()
        >>>         self.conv1 = nn.Sequential(
        >>>             nn.Conv2d(
        >>>                 in_channels=1,
        >>>                 out_channels=16,
        >>>                 kernel_size=5,
        >>>                 stride=1,
        >>>                 padding=2,
        >>>             ),
        >>>             nn.ReLU(),
        >>>             nn.MaxPool2d(kernel_size=2),
        >>>             nn.Conv2d(in_channels=16,
        >>>                 out_channels=16,
        >>>                 kernel_size=5,
        >>>                 stride=1,
        >>>                 padding=2,)
        >>>         )
        >>>         self.conv2 = nn.Sequential(
        >>>             nn.Conv2d(16, 32, 5, 1, 2),
        >>>             nn.ReLU(),
        >>>             nn.MaxPool2d(2),
        >>>         )
        >>>         self.out = nn.Linear(32 * 7 * 7, 10)
        >>>         linger.disable_quant(self.out)
    Notes:
        disable_quant 应该在linger.init函数之前调用才生效

        """
    queue = [module]
    while len(queue) > 0:
        node = queue.pop(0)
        if type(node) in SupportQuantTorchModules:
            _SingletonContainCustomModules()._register(node, None)
        for _, submodule in node.named_children():
            queue.append(submodule)


def quant_module(module: nn.Module, type_modules: Tuple = DefaultQuantIntXOP, mode: QuantMode = QuantMode.QValue, data_bits: int = 8, parameter_bits: int = 8, out_bits: int = None):
    r"""对module进行自定义量化设置,这是一个通用的量化策略设置接口,该接口支持inline使用

        Args:
            module: 需要量化的模型或者模块
            type_modules(tuple):量化针对的类型,默认为(nn.Linear, nn.Conv2d, nn.ConvTranspose2d, nn.GRU, nn.LSTM)
            mode(QuantMode):量化模式,默认为Q值量化
            data_bits(int):推理过程中data(激活值)量化精度bit数，默认为8bit
            parameter_bits(int):权重的量化精度bit数，默认为8bit
            out_bits(int or None):输出数值的精度，默认为None，表示float精度 
        .. math:: 
            \text {save_weight_bits}=\begin{cases}
                8 & parameter\_bits\leq 8 \\
                16 & 8 < parameter\_bits\leq 16 \\
                32 & parameter\_bits \geq 16
            \end{cases}
        Notes:
            quant_module 应该在linger.init函数之前调用才生效
            如果module weights中有bias, 如果onnx 导出, 存储使用32bit
        """
    if not isinstance(type_modules, tuple):
        type_modules = (type_modules,)
    for t in type_modules:
        if t not in SupportQuantTorchModules:      # type_modules is a subset of SupportQuantTorchModules
            logger.fatal(str(t)+" is not supprt quant in linger now")
            exit(-1)

    queue = [module]
    while len(queue) > 0:
        node = queue.pop(0)
        if type(node) in type_modules and type(node) in SupportQuantTorchModules:      # and 后 判断多余
            qinfo = QuantInfo()
            qinfo.set_data_bits(data_bits)
            qinfo.set_parameter_bits(parameter_bits)
            qinfo.set_output_bits(out_bits)
            qinfo.set_mode(mode)
            _SingletonContainCustomModules()._register(node, qinfo)
        for _, submodule in node.named_children():
            queue.append(submodule)


def quant_module_by_type(module: nn.Module, type_modules: Tuple = DefaultQuantIntXOP, mode: QuantMode = QuantMode.QValue, data_bits: int = 8, parameter_bits: int = 8, out_bits: int = None):
    r"""对module进行自定义量化设置,包括激活值，weight以及输出激活值都是16bit,这是一个通用的量化策略设置接口,该接口支持inline使用

        Args:
            module: 需要量化的模型或者模块
            type_modules(tuple):量化针对的类型,默认为(nn.Linear, nn.Conv2d, nn.ConvTranspose2d, nn.GRU, nn.LSTM)
            mode(QuantMode):量化模式,默认为Q值量化
            data_bits(int):推理过程中data(激活值)量化精度bit数，默认为8bit
            parameter_bits(int):权重的量化精度bit数，默认为8bit
            out_bits(int or None):输出数值的精度，默认为None，表示float精度 

        Notes:
            quant_module_by_type 应该在linger.init函数之前调用才生效
            如果module weights中有bias, 如果onnx 导出, 存储使用32bit
        """
    if type(type_modules) is not tuple:
        type_modules = (type_modules,)
    for user_module_type in type_modules:
        assert user_module_type in SupportQuantTorchModules, 'currently not support quant of ' + \
            str(user_module_type)
    queue = [module]
    while len(queue) > 0:
        node = queue.pop(0)
        for t in type_modules:
            if type(node) == t:
                qinfo = QuantInfo()
                qinfo.set_data_bits(data_bits)
                qinfo.set_parameter_bits(parameter_bits)
                qinfo.set_output_bits(out_bits)
                qinfo.set_mode(mode)
                _SingletonContainCustomModules()._register(node, qinfo)
        for _, submodule in node.named_children():
            queue.append(submodule)


def quant_tensor(module: nn.Module, x: torch.Tensor, name: str = '_default_layername', mode: QuantMode = QuantMode.QValue, bits: int = 8, zero_point: int = 0) -> torch.Tensor:
    r"""对tensor进行量化

    Args:
        module(torch.nn.Module):tensor 量化所在的module，如果是在forward代码里面，一般是self
        x(torch.Tensor):量化tensor的tensor
        name(str):量化后module名字，如果同一forward中出现多个需要量化的tensor，该名字应设置成不一样，默认为'_default_layername'
        mode(QuantMode):量化模式，默认QValue
        bits(int):tensor量化的bit数，默认为8bit
    Returns:
        返回量化后的x(tensor)
    Example:
        >>> def forward(self, x):        
        >>>     x = self.conv1(x)
        >>>     x = linger.quant_tensor(self,x,'i_am_the_key_code_line')
        >>>     x = self.conv2(x)
        >>>     x = x.view(x.size(0), -1)           
        >>>     output = self.out(x)
        >>> return output

    """
    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + '_iq_tensor_quant_' + name
    if hasattr(module, var_name):
        round_layer = getattr(module, var_name)
    else:
        round_layer = ScaledRoundLayer(
            mode=mode, bits=bits, zero_point=zero_point)
        round_layer.training = module.training
        round_layer = round_layer.to(x.device)
        setattr(module, var_name, round_layer)
    x = round_layer(x)
    return x


def hook_pre_pack_forward(module, input):
    input = list(input)
    if isinstance(input[0], tuple):
        orig_input = input[0]
        input_ori, lengths, batch_first, enforce_sorted = orig_input
        packed_input = torch_pack_padded_sequence(
            input_ori, lengths, batch_first, enforce_sorted)
        input[0] = packed_input
    return tuple(input)


def hook_pre_pad_forward(module, input, output):
    if isinstance(input[0], tuple):
        output_packed = output[0]
        output_unpack, lengths = torch_pad_packed_sequence(
            output_packed, module.batch_first)
        return (output_unpack, lengths), output[1]
    else:
        return output


def init(model: nn.Module, *, quant_modules: Tuple = DefaultQuantIntXOP, parameter_bits: int = 8, mode: QuantMode = QuantMode.QValue) -> nn.Module:
    data_bits = 8
    out_bits = 8
    assert data_bits > 0 and data_bits <= 32, "data_bits should between 0 and 32"
    assert parameter_bits > 0 and parameter_bits <= 32, "parameter_bits should between 0 and 32"
    assert data_bits + parameter_bits <= 32, "data_bits + parameter_bits less than 32"
    assert out_bits is None or out_bits > 0 and out_bits <= 32, "out_bits should between 0 and 32"
    if type(quant_modules) is not tuple:
        quant_modules = (quant_modules,)
    quant_modules = set(list(quant_modules) + [NormalizeConvBN1d, NormalizeConvBN2d, NormalizeConv2d, NormalizeConv1d, NormalizeConvTranspose2d,
                        NormalizeLinear, nn.ReLU6, NormalizeFastGRU, NormalizeFastLSTM, NormalizeBatchNorm2d, NormalizeEmbedding, NormalizeLayerNorm])
    for user_module_type in quant_modules:
        assert user_module_type in SupportQuantTorchModules, 'currently not support quant of ' + \
            str(user_module_type)

    if config.IQTensor.iqcat:
        torch.cat = linger_cat
    if config.IQTensor.iqclamp:
        torch.clamp = linger_clamp
        torch.clamp_ = linger_clamp_
    if config.IQTensor.iqsigmoid:
        torch.sigmoid = linger_sigmoid
        torch.sigmoid_ = linger_sigmoid_
    if config.IQTensor.iqtanh:
        torch.tanh = linger_tanh
        torch.tanh_ = linger_tanh_
    if config.IQTensor.softmaxint:
        torch.softmax = linger_softmax
        torch.nn.functional.softmax = linger_softmax
    if config.IQTensor.logsoftmaxint:
        torch.log_softmax = linger_logsoftmax
        torch.nn.functional.log_softmax = linger_logsoftmax
    if config.IQTensor.iqvar:
        torch.var = linger_var
    if config.FunctionQuant.bmm:
        torch.bmm = linger_bmm
    if config.FunctionQuant.channel_shuffle:
        linger.channel_shuffle = channel_shuffle_quant
    torch.nn.functional.dropout = linger_dropout
    torch.onnx.export = linger_export
    if nn.LSTM in quant_modules or nn.GRU in quant_modules:
        torch.nn.utils.rnn.pack_padded_sequence = linger_pack_padded_sequence
        torch.nn.utils.rnn.pad_packed_sequence = linger_pad_packed_sequence
    device = get_device(model)
    queue = [model]
    while len(queue) > 0:
        node = queue.pop(0)
        # add for get father module in current forward
        node.register_forward_pre_hook(hook_pre_forward)
        node.register_forward_hook(hook_forward)
        setattr(node, LINGER_MODE, mode)
        setattr(node, LINGER_OBIT, out_bits)
        for name, submodule in node.named_children():
            # customed by user, set Singleton Class for get module register_keys by class instance
            if _SingletonContainCustomModules()._is_registered(submodule):
                quant_info = _SingletonContainCustomModules().get(submodule)
                if quant_info == None:  # disabled module
                    if isinstance(submodule, (nn.LSTM, nn.GRU, NormalizeFastGRU, NormalizeFastLSTM)):
                        submodule.register_forward_pre_hook(
                            hook_pre_pack_forward)
                        submodule.register_forward_hook(hook_pre_pad_forward)
                    continue
                else:
                    r_module = _replaceOp(submodule, quant_info.mode, quant_info.data_bits,
                                          quant_info.parameter_bits, quant_info.output_bits)
                    assert r_module is not None
                    setattr(node, name, r_module)
            elif type(submodule) in SupportQuantTorchModules and type(submodule) in quant_modules:  # not customed
                r_module = _replaceOp(
                    submodule, mode, data_bits, parameter_bits, out_bits)
                assert r_module is not None
                setattr(node, name, r_module)
            else:
                queue.append(submodule)

    def quant_tensor_pre_hook(state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):

        def quant_tensor_layer(module, prefix=''):
            local_name_params = itertools.chain(
                module._parameters.items(), module._buffers.items())
            local_state = {k: v for k, v in local_name_params if v is not None}
            for key in state_dict.keys():
                if LINGER_MIX_INT8_MANUAL_ROUND_LAYERS in key:
                    if key.startswith(prefix):
                        full_input_name = key[len(prefix):]
                        # get the name of param/buffer/child
                        input_name = full_input_name.split('.', 1)[0]
                        if input_name not in module._modules and input_name not in local_state:
                            if '_iqadd_' in input_name:
                                iq_layer = iqAddLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqmul_' in input_name:
                                iq_layer = iqMulLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqcat_' in input_name:
                                iq_layer = iqCatLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqsigmoid_' in input_name:
                                iq_layer = iqSigmoidLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqtanh_' in input_name:
                                iq_layer = iqTanhLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqclamp_' in input_name:
                                iq_layer = iqClampLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqdiv_' in input_name:
                                iq_layer = iqDivLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqsum_' in input_name:
                                iq_layer = iqSumLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iqvar_' in input_name:
                                iq_layer = iqVarLayer()
                                iq_layer.training = model.training
                                iq_layer = iq_layer.to(device)
                                setattr(module, input_name, iq_layer)
                            elif '_iq_tensor_quant_' in input_name:
                                round_layer = ScaledRoundLayer(
                                    mode=mode, bits=data_bits)
                                round_layer.training = model.training
                                round_layer = round_layer.to(device)
                                setattr(module, input_name, round_layer)
                            elif '_function_bmm_' in input_name:
                                bmm_layer = BmmInt(data_bits=8, mode=mode,)
                                bmm_layer.training = model.training
                                bmm_layer = bmm_layer.to(device)
                                setattr(module, input_name, bmm_layer)
                            elif '_SoftmaxInt_' in input_name:
                                softmax_layer = softmaxIntLayer(
                                    data_bits=8, mode=mode,)
                                softmax_layer.training = model.training
                                softmax_layer = softmax_layer.to(device)
                                setattr(module, input_name, softmax_layer)

            for name, children in module._modules.items():
                if children is not None:
                    quant_tensor_layer(children, prefix + name + '.')

        quant_tensor_layer(model, prefix)
        quant_tensor_layer = None

    model._register_load_state_dict_pre_hook(quant_tensor_pre_hook)
    if model.training:
        model.train()
    else:
        model.eval()
    model.to(device)
    return model



from typing import Tuple, TypeVar, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..quant import NormalizeFunction

T = TypeVar('T')
_scalar_or_tuple_2_t = Union[T, Tuple[T, T]]
_size_2_t = _scalar_or_tuple_2_t[int]


class NormalizeConvTranspose2d(nn.ConvTranspose2d):
    def __init__(self, in_channels: int,
                 out_channels: int,
                 kernel_size: _size_2_t,
                 stride: _size_2_t = 1,
                 padding: _size_2_t = 0,
                 output_padding: _size_2_t = 0,
                 groups: int = 1,
                 bias: bool = True,
                 dilation: int = 1,
                 padding_mode: str = 'zeros',
                 normalize_data=None, normalize_weight=None, normalize_bias=None,
                 ) -> None:
        assert normalize_data is None or normalize_data > 0, 'normalize value is None or must >0'
        assert normalize_weight is None or normalize_weight > 0, 'normalize value is None or must >0'
        assert normalize_bias is None or normalize_bias > 0, 'normalize value is None or must >0'

        # venus limits
        if type(stride) == int:
            if stride == 2:
                assert kernel_size in (
                    2, 3, 4, 5), f"in NormalizeConvTranspose2d op, when stride_h == 2, kernel_h must be 2, 3, 4 or 5, but you have kernel_size: {kernel_size}"
            elif stride == 4:
                assert kernel_size in (
                    4, 5), f"in NormalizeConvTranspose2d op, when stride_h == 4, kernel_h must be 4 or 5, but you have kernel_size: {kernel_size}"
        else:
            if stride[1] == 2:
                assert kernel_size[1] in (
                    2, 3, 4, 5), f"in NormalizeConvTranspose2d op, when stride_h == 2, kernel_h must be 2, 3, 4 or 5, but you have kernel_size[1]: {kernel_size[1]}"
            if stride[1] == 4:
                assert kernel_size[1] in (
                    4, 5), f"in NormalizeConvTranspose2d op, when stride_h == 4, kernel_h must be 4 or 5, but you have kernel_size[1]: {kernel_size[1]}"
            if stride[0] == 2:
                assert kernel_size[0] in (
                    2, 3, 4, 5), f"in NormalizeConvTranspose2d op, when stride_2 == 2, kernel_2 must be 2, 3, 4 or 5, but you have kernel_size[0]: {kernel_size[0]}"
            if stride[0] == 4:
                assert kernel_size[0] in (
                    4, 5), f"in NormalizeConvTranspose2d op, when stride_w == 4, kernel_w must be 4 or 5, but you have kernel_size[0]: {kernel_size[0]}"

        nn.ConvTranspose2d.__init__(self, in_channels, out_channels, kernel_size,
                                    stride, padding, output_padding, groups, bias, dilation, padding_mode)
        self.normalize_data = normalize_data
        self.normalize_weight = normalize_weight
        self.normalize_bias = normalize_bias

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        normalized_weight = self.weight
        if self.normalize_weight:
            normalized_weight = NormalizeFunction.apply(
                normalized_weight, self.normalize_weight, self.training)
        normalized_bias = self.bias
        if (self.bias is not None) and (self.normalize_bias is not None):
            normalized_bias = NormalizeFunction.apply(
                normalized_bias, self.normalize_bias, self.training)
        out = F.conv_transpose2d(
            input, normalized_weight, normalized_bias, self.stride, self.padding,
            self.output_padding, self.groups, self.dilation)
        if self.normalize_data is not None:
            out = NormalizeFunction.apply(
                out, self.normalize_data, self.training, False)
        return out

    def extra_repr(self):
        s = nn.ConvTranspose2d.extra_repr(self)
        extra_s = ',normalize_data:{normalize_data},normalize_weight:{normalize_weight},normalize_bias:{normalize_bias}'.format(
            **self.__dict__)
        return s+extra_s


__all__ = ['NormalizeConvTranspose2d']



import linger
import numpy
import torch
import torch.nn as nn
from linger import NormalizeFastGRU


def test_lstmpint_net():

    def getacc(lprob, target):
        num_class = lprob.size()[1]
        _, new_target = torch.broadcast_tensors(lprob, target)

        remove_pad_mask = new_target.ne(-1)
        lprob = lprob[remove_pad_mask]

        target = target[target != -1]
        target = target.unsqueeze(-1)

        lprob = lprob.reshape((-1, num_class))

        preds = torch.argmax(lprob, dim=1)

        correct_holder = torch.eq(preds.squeeze(), target.squeeze()).float()

        num_corr = correct_holder.sum()
        num_sample = torch.numel(correct_holder)
        acc = num_corr/num_sample
        return acc

    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv0 = nn.Conv2d(1, 100, kernel_size=(
                1, 3), padding=(0, 1), groups=1, bias=True)
            self.bn0 = nn.BatchNorm2d(100)
            self.relu0 = nn.ReLU()
            self.conv1 = nn.Conv2d(100, 100, kernel_size=(
                1, 3), padding=(0, 1), groups=1, bias=True)
            self.bn1 = nn.BatchNorm2d(100)
            self.relu1 = nn.ReLU()
            self.lstmp = nn.GRU(100, 50, num_layers=1,
                                batch_first=True, bidirectional=True)
            self.final_conv = nn.Conv2d(100, 10, 1, 1, 0)

        def forward(self, input, batch_lengths=None, initial_state=None):
            x = self.conv0(input)
            x = self.bn0(x)
            x = self.relu0(x)
            x = self.conv1(x)
            x = self.bn1(x)
            x = self.relu1(x)
            n, c, h, w = x.shape
            x = x.reshape((n, -1, 1, w)).squeeze(2)
            x = x.permute((0, 2, 1))  # b t d
            x = nn.utils.rnn.pack_padded_sequence(
                x, batch_lengths, batch_first=True, enforce_sorted=False)
            x, _ = self.lstmp(x, initial_state)  # output b, t, h (10, 10, 100)
            x, _ = nn.utils.rnn.pad_packed_sequence(x, batch_first=True)
            x = x.permute((2, 0, 1))
            d, b, t = x.shape
            x = x.reshape((1, d, 1, b*t))  # (1, 100, 1, 100)
            x = self.final_conv(x)  # (1, 10, 1, 100) (d, b*t)
            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    numpy.random.seed(1)
    dummy_input = torch.randn(10, 1, 1, 10).cuda()
    label = torch.randint(10, (10, 10)).cuda()  # class=10
    mask = torch.ones(10, 10)
    for i in range(9):
        index = numpy.random.randint(5, 10)
        mask[i, index:] = 0
        label[i, index:] = -1

    input_lengths = mask.long().sum(1).cpu().numpy()
    input_lengths = torch.tensor(input_lengths)  # .cuda()
    batch_size = 10
    hidden_size = 50
    size = 2
    initial_state = torch.zeros(size, batch_size, hidden_size).cuda()
    net = Net().cuda()
    replace_modules = (nn.Conv2d, nn.GRU, nn.BatchNorm2d)
    criterion = nn.CrossEntropyLoss(ignore_index=-1)
    # net = linger.normalize_layers(net)

    optimizer = torch.optim.Adam(net.parameters(), lr=0.001)

    print('net: ', net)
    # net.load_state_dict(torch.load('data.ignore/gru.pt'))
    loss = None
    for i in range(5):
        optimizer.zero_grad()
        out = net(dummy_input, input_lengths, initial_state)
        out = out.squeeze().permute((1, 0))  # (b*t, d)
        loss = criterion(out, label.reshape(-1))
        if i % 50 == 0:
            print('loss: ', loss)
            acc = getacc(out, label.reshape(-1, 1))
            print('train acc: ', acc)
        loss.backward()
        optimizer.step()

    net.eval()
    out1 = net(dummy_input, input_lengths, initial_state)
    out1 = out1.squeeze().permute((1, 0))
    acc = getacc(out1, label.reshape(-1, 1))
    print('test acc: ', acc)
    torch.save(net.state_dict(), 'data.ignore/gru.pt')
    # out3 = net(dummy_input, input_lengths, initial_state)
    # out3 = out3.squeeze().permute((1,0))
    # input_lengths = torch.tensor(input_lengths)#.cuda()
    # with torch.no_grad():
    #     net.eval()
    #     torch.onnx.export(net, (dummy_input, input_lengths, initial_state), "data.ignore/gru.onnx", export_params=True,
    #                       opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    out2 = net(dummy_input, input_lengths, initial_state)
    out2 = out2.squeeze().permute((1, 0))
    # print('out: ', out1)
    assert (out1 == out2).all()


def test_normalize_gru():

    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.rnn = nn.GRU(10, 20, 1, batch_first=True, bidirectional=True)
            # self.rnn = NormalizeFastGRU(10, 20, 1, batch_first=True, bidirectional=True)

        def forward(self, input, h0):
            x, hn = self.rnn(input, h0)

            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    numpy.random.seed(1)

    dummy_input = torch.randn(3, 5, 10).cuda()
    h0 = torch.randn(2, 3, 20).cuda()

    model = Net().cuda()

    model = linger.normalize_layers(model)

    # torch.save(model.state_dict(), "data.ignore/torch_gru.pt")
    # model.load_state_dict(torch.load("data.ignore/torch_gru.pt"))
    model.eval()

    out1 = model(dummy_input, h0)

    print('model: ', model)

    with torch.no_grad():
        model.eval()
        torch.onnx.export(model, (dummy_input, h0), "data.ignore/normalize_torch_gru.onnx", export_params=True,
                          opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import torch
import torch.nn as nn
import linger
import numpy as np


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                              padding=1, bias=True)
        self.bn = nn.BatchNorm2d(10)
        self.relu = nn.ReLU()
        self.conv1 = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                               padding=1, bias=True)
        self.bn1 = nn.BatchNorm2d(10)
        self.relu1 = nn.ReLU()
        self.conv2 = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                               padding=1, bias=True)
        self.bn2 = nn.BatchNorm2d(10)
        self.relu2 = nn.ReLU()
        self.fc = nn.Linear(1000, 100)

    def forward(self, input):
        x = self.conv(input)
        x = self.bn(x)
        x1 = self.relu(x)

        x = self.conv1(input)
        x2 = self.bn1(x)
        x2 = self.relu(x2)
        # y = linger.from_torch_tensor(torch.tensor(1),0.5,8).cuda()
        x2 = x2 + 1
        # x2 = self.relu1(x)
        x = x1 / 2
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu2(x)
        n, c, h, w = x.shape
        x = x.view((n, c*h*w))
        x = self.fc(x)
        return x


def test_conv_linear():

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    # random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    aa = torch.randn(1, 10, 10, 10).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.Linear, nn.BatchNorm2d)
    net.train()
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    linger.FuseConvBNAheadRelu(
        net, aa, fused_bn=False, ahead_bn_relu=True, ahead_conv_relu=True)
    net = linger.init(net, quant_modules=replace_tuple,
                      mode=linger.QuantMode.QValue)
    print(net)
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)
    loss = None
    for i in range(200):
        optimizer.zero_grad()
        out = net(aa)
        loss = criterion(out, target)
        if i % 20 == 0:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()
    # assert loss < 1e-12, 'training loss error'
    net.eval()
    torch.save(net.state_dict(), 'data.ignore/conv_iqdiv.pt')
    print(net)
    out1 = net(aa)
    # print(out1)
    with torch.no_grad():
        torch.onnx.export(net, aa, "data.ignore/conv_iqdiv1.onnx", export_params=True,
                          opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    assert abs(out1.mean() - 1) < 0.1






def notest_clamp_gru_in_onnxinfer():
    pass

    # import onnxinfer
    # import torch
    # import numpy
    # torch.manual_seed(1)
    # torch.cuda.manual_seed_all(1)
    # numpy.random.seed(1)

    # dummy_input = torch.randn(3, 5, 10).cuda()
    # h0 = torch.randn(2, 3, 20).cuda()

    # sessoption = onnxinfer.InferSessionOptions()
    # sess = onnxinfer.InferSession(
    #     'data.ignore/normalize_torch_gru.onnx', sessoption, is_fuse=False, save_transform_model=None)
    # data = {sess.GetInputNames()[0]: dummy_input, sess.GetInputNames()[1]: h0}
    # rlt = sess.Run(data_in=data)
    # output = rlt[0].AsReadOnlyNumpy()

    # sessoption1 = onnxinfer.InferSessionOptions()
    # sess1 = onnxinfer.InferSession(
    #     'data.ignore/torch_gru.onnx', sessoption1, is_fuse=False, save_transform_model=None)
    # data1 = {sess1.GetInputNames()[0]: dummy_input,
    #          sess1.GetInputNames()[1]: h0}
    # rlt1 = sess1.Run(data_in=data1)
    # output1 = rlt1[0].AsReadOnlyNumpy()



from .avgpool2d_int import AvgPool2dInt
from .batchnorm_int import BatchNormInt
from .layernorm_int import LayerNormInt
from .bmm_int import BmmInt
from .conv1d_int import Conv1dInt
from .conv_int import Conv2dInt
from .convtranspose_int import ConvTranspose2dInt
from .embedding_int import EmbeddingInt
from .gru_int import GRUInt
from .iqtensor import *
from .linear_int import LinearInt
from .linger_functional import *
from .lstm_int import LSTMInt
from .module_self import *
from .ops import ModuleIntConfig
from .ops_configs import (DefaultQuantIntXOP, SupportQuantedIntModules,
                          SupportQuantTorchModules)
from .ops_names import *
from .relu6_int import ReLU6Int
from .requant import Requant
from .scaledround_int import ScaledRoundLayer



import torch
from linger.ops import from_torch_tensor


def test_relu():
    a = torch.randn((16, 16), requires_grad=True)
    x = a.detach().clone()
    x.requires_grad_()
    b = torch.relu(a)
    b.sum().backward()

    y = from_torch_tensor(x, 5, 8)
    c = torch.relu(y)
    assert c.scale_data == 5
    assert c.bits == 8
    c.sum().backward()

    assert (a.grad - x.grad).sum() == 0


def test_relu_():
    a = torch.randn((16, 16), requires_grad=True)
    x = a.detach().clone()
    x.requires_grad_()
    b = torch.relu_(a)
    b.sum().backward()

    y = from_torch_tensor(x, 5, 8)
    c = torch.relu_(y)
    assert c.scale_data == 5
    assert c.bits == 8
    c.sum().backward()

    assert (a.grad - x.grad).sum() == 0


def test_max_pool2d():
    a = torch.randn((1, 16, 16), requires_grad=True)
    x = a.detach().clone()
    x.requires_grad_()
    b = torch.max_pool2d(a, kernel_size=(2, 2), stride=2, padding=0)
    b.sum().backward()

    y = from_torch_tensor(x, 5, 8)
    c = torch.max_pool2d(y, kernel_size=(2, 2), stride=2, padding=0)
    assert c.scale_data == 5
    assert c.bits == 8
    c.sum().backward()

    assert (a.grad - x.grad).sum() == 0
